# -*- coding: utf-8 -*-


# Built-in
import warnings
import itertools as itt


# Common
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from mpl_toolkits.mplot3d import Axes3D
import datastock as ds


# specific


# ###############################################################
# ###############################################################
#                       Compute
# ###############################################################


def sinogram(
    coll=None,
    key=None,
    segment=None,
    # config
    config=None,
    kVes=None,
    # sinogram ref point
    R0=None,
    Z0=None,
    # sinogram options
    ang=None,
    ang_units=None,
    impact_pos=None,
    pmax=None,
    Dphimax=None,
    # plotting options
    plot=None,
    color=None,
    marker=None,
    label=None,
    sketch=None,
    # other options
    verb=None,
    # figure
    dax=None,
    dmargin=None,
    fs=None,
    wintit=None,
):

    # -----------------
    # check
    # -----------------

    (
        key,
        segment,
        # config
        config,
        kVes,
        # sinogram options
        ang,
        ang_units,
        impact_pos,
        # plotting options
        plot,
        # other options
        verb,
    ) = _check(**locals())

    if len(key) == 0 and config is None:
        print('None', key, config)
        return

    # -----------------
    # get R0, Z0
    # -----------------

    R0, Z0, Dphimax = _get_RZ0(R0=R0, Z0=Z0, config=config, kVes=kVes)

    # -----------------
    # compute for rays
    # -----------------

    dout = {}
    dfail = {}
    for k0 in key:
        dout[k0] = _compute_rays(
            coll=coll,
            kray=k0,
            segment=segment,
            R0=R0,
            Z0=Z0,
            Dphimax=Dphimax,
            # options
            verb=verb,
        )

        nans = np.isnan(dout[k0]['ang']).sum()
        if nans == dout[k0]['ang'].size:
            dfail[k0] = 'all nans'
        elif nans > 0:
            dfail[k0] = f"{nans} / {dout[k0]['ang'].size} nans"

    # warnings
    if len(dfail) > 0 and verb == 1:
        lstr = [f"\t- {k0}: {v0}" for k0, v0 in dfail.items()]
        msg = (
            "Some rays seem to have no impact solution:\n"
            + "\n".join(lstr)
        )
        warnings.warn(msg)

    # -----------------
    # compute for config
    # -----------------

    dout_config = None
    if config is not None:
        dout_config = _compute_config(
            config=config,
            kVes=kVes,
            R0=R0,
            Z0=Z0,
        )

    # ------------
    # adjust
    # ------------

    # convert impact parameter
    if impact_pos is False:
        for k0, v0 in dout.items():
            ineg = v0['ang'] < 0.
            dout[k0]['impact'][ineg] = -v0['impact'][ineg]
            dout[k0]['ang'][ineg] = np.arctan2(
                np.sin(v0['ang'][ineg] + np.pi),
                np.cos(v0['ang'][ineg] + np.pi),
            )

        if dout_config is not None:
            ineg = dout_config['ang'] < 0.
            dout_config['pmax'][ineg] = -dout_config['pmax'][ineg]
            dout_config['ang'][ineg] = np.arctan2(
                np.sin(dout_config['ang'][ineg] + np.pi),
                np.cos(dout_config['ang'][ineg] + np.pi),
            )

    # -------------
    # convert angle

    if ang == 'xi':
        for k0, v0 in dout.items():
            dout[k0]['ang'] = np.arctan2(
                np.sin(v0['ang'] - np.pi/2.),
                np.cos(v0['ang'] - np.pi/2.),
            )

        if dout_config is not None:
            dout_config['ang'] = np.arctan2(
                np.sin(dout_config['ang'] - np.pi/2),
                np.cos(dout_config['ang'] - np.pi/2),
            )

    elif ang == 'zeta':
        for k0, v0 in dout.items():
            zeta = np.arctan2(
                np.sin(np.pi - v0['ang']),
                np.cos(np.pi - v0['ang']),
            )
            if impact_pos is False:
                zeta[zeta > np.pi/2] = zeta[zeta > np.pi/2] - np.pi
            dout[k0]['ang'] = zeta

        if dout_config is not None:
            zeta = np.arctan2(
                np.sin(np.pi - dout_config['ang']),
                np.cos(np.pi - dout_config['ang']),
            )
            if impact_pos is False:
                zeta[zeta > np.pi/2] = zeta[zeta > np.pi/2] - np.pi
            dout_config['ang'] = zeta

    # -------------------
    # convert angle units

    if ang_units == 'deg':
        for k0, v0 in dout.items():
            dout[k0]['ang'] = v0['ang'] * 180 / np.pi

        if dout_config is not None:
            dout_config['ang'] = dout_config['ang'] * 180 / np.pi

    # -------------
    # store in dict

    for k0, v0 in dout.items():
        dout[k0]['ang_var'] = ang
        dout[k0]['ang_units'] = ang_units
        dout[k0]['impact_pos'] = impact_pos

    if dout_config is not None:
        dout_config['ang_var'] = ang
        dout_config['ang_units'] = ang_units
        dout_config['impact_pos'] = impact_pos

    # ------------
    # plot
    # ------------

    if plot is True:

        dax = _plot(
            dout=dout,
            dout_config=dout_config,
            # sinogram ref point
            R0=R0,
            Z0=Z0,
            # sinogram options
            ang=ang,
            ang_units=ang_units,
            impact_pos=impact_pos,
            pmax=pmax,
            # plotting options
            color=color,
            marker=marker,
            label=label,
            sketch=sketch,
            # figure
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            wintit=wintit,
        )

        return dout, dax

    return dout


# ###############################################################
# ###############################################################
#                       Check
# ###############################################################


def _check(
    coll=None,
    key=None,
    segment=None,
    # config
    config=None,
    kVes=None,
    # sinogram ref point
    R0=None,
    Z0=None,
    # sinogram options
    ang=None,
    ang_units=None,
    impact_pos=None,
    # plotting options
    plot=None,
    # other options
    verb=None,
    # unused
    **kwdargs,
):

    # -------------
    # key
    # -------------

    # allowed
    lrays = list(coll.dobj.get('rays', {}).keys())
    ldiag = [
        k0 for k0, v0 in coll.dobj.get('diagnostic', {}).items()
        if any([v1.get('los') is not None for v1 in v0['doptics'].values()])
    ]
    ldiag = list(coll.dobj.get('diagnostic', {}).keys())

    if isinstance(key, str):
        key = [key]

    # check
    key = ds._generic_check._check_var_iter(
        key, 'key',
        types=list,
        types_iter=str,
        allowed=ldiag + lrays,
    )

    # convert to list of rays
    if any([k0 in ldiag and k0 not in lrays for k0 in key]):
        for ii, k0 in enumerate(key):
            if k0 in lrays:
                key[ii] = [k0]
            else:
                lcam = coll.dobj['diagnostic'][k0]['camera']
                doptics = coll.dobj['diagnostic'][k0]['doptics']
                key[ii] = [
                    doptics[k1]['los'] for k1 in lcam
                    if doptics[k1]['los'] is not None
                ]

        if any([isinstance(k0, list) for k0 in key]):
            key = list(itt.chain.from_iterable(key))

    # ------------
    # segment

    segment = ds._generic_check._check_var(
        segment, 'segment',
        types=int,
        default=-1,
    )

    # -------------------
    # config

    if config is not None:
        lkVes = list(config.dStruct['dObj']['Ves'].keys())
        if kVes is None:
            lkVes = list(config.dStruct['dObj']['Ves'].keys())
            ls = [
                config.dStruct['dObj']['Ves'][k0].dgeom['Surf']
                for k0 in lkVes
            ]
            kVes = lkVes[np.argmin(ls)]

        kVes = ds._generic_check._check_var(
            kVes, 'kVes',
            types=str,
            allowed=lkVes,
        )

    # ------------
    # ang

    ang = ds._generic_check._check_var(
        ang, 'ang',
        types=str,
        default='theta',
        allowed=['xi', 'theta', 'zeta'],
    )

    # ------------
    # ang_units

    ang_units = ds._generic_check._check_var(
        ang_units, 'ang_units',
        types=str,
        default='deg',
        allowed=['deg', 'radian'],
    )

    # ------------
    # impact_pos

    impact_pos = ds._generic_check._check_var(
        impact_pos, 'impact_pos',
        types=bool,
        default=True,
    )

    # ------------
    # plot

    plot = ds._generic_check._check_var(
        plot, 'plot',
        types=bool,
        default=True,
    )

    # --------------
    # verb

    if isinstance(verb, bool):
        verb = 2 if verb else 0

    verb = ds._generic_check._check_var(
        verb, 'verb',
        types=int,
        default=2,
        allowed=[0, 1, 2],
    )

    return (
        key,
        segment,
        # config
        config,
        kVes,
        # sinogram options
        ang,
        ang_units,
        impact_pos,
        # plotting options
        plot,
        # other options
        verb,
    )


def _get_RZ0(
    R0=None,
    Z0=None,
    Dphimax=None,
    config=None,
    kVes=None,
):

    # ------------
    # R0

    if R0 is None:
        if config is None:
            msg = "Please provide a value for R0!"
            raise Exception(msg)

        R0 = config.dStruct['dObj']['Ves'][kVes].dgeom['BaryS'][0]

    # ------------
    # Z0

    if Z0 is None:
        if config is None:
            Z0 = 0.
        else:
            Z0 = config.dStruct['dObj']['Ves'][kVes].dgeom['BaryS'][1]

    # ---------------
    # Dphimax = max authorized Delta phi

    Dphimax = ds._generic_check._check_var(
        Dphimax, 'Dphimax',
        types=(int, float),
        sign='>0',
        default=3*np.pi/4,
    )

    return R0, Z0, Dphimax

# ###############################################################
# ###############################################################
#               Compute - rays
# ###############################################################


def _compute_rays(
    coll=None,
    kray=None,
    segment=None,
    R0=None,
    Z0=None,
    Dphimax=None,
    # options
    verb=None,
):

    # --------------------------------
    # compute (for first segment only)
    # --------------------------------

    # segment
    seg_pts = segment
    seg_vect = segment
    if segment < 0:
        seg_pts -= 1
    end_pts = seg_pts + 1

    # starting points
    ptsx, ptsy, ptsz = coll.get_rays_pts(kray)

    endx, endy, endz = [pp[end_pts, ...] for pp in [ptsx, ptsy, ptsz]]
    ptsx, ptsy, ptsz = [pp[seg_pts, ...] for pp in [ptsx, ptsy, ptsz]]
    length = np.sqrt(
        (endx - ptsx)**2
        + (endy - ptsy)**2
        + (endz - ptsz)**2
    )

    # unit vectors
    vectx, vecty, vectz = coll.get_rays_vect(kray, norm=True)
    vectx, vecty, vectz = [vv[seg_vect, ...] for vv in [vectx, vecty, vectz]]
    vnorm2d = np.sqrt(vectx**2 + vecty**2)

    # dphi
    phi = np.arctan2(ptsy, ptsx)
    cosdphi = np.cos(phi) * vectx / vnorm2d + np.sin(phi) * vecty / vnorm2d
    dphi = np.arccos(np.abs(cosdphi))

    # ----------------------
    # solve
    # ----------------------

    # ---------
    # statement

    # O = origin
    # A = origin of LOS
    #     OA = ZA ez + RA eRA
    # M = point on LOS
    #     AM = k v
    #         v = vz + vpar
    #         Z = ZA + k vz
    #         R^2 = RA^2 + k^2 vpar^2 + 2 k RA (v.eRA)
    # C = point on z-axis at Z0
    #     OC = Z0 ez
    # T = point on circle where tangency happens
    #     CT = R0 eR0

    # Introduce impact factor p and unit vector er /
    #     TM = p er

    # Exception: M on axis (C not defined)

    # solve k /
    #     AM = k v
    #     AM.MT = 0

    # ---------
    # method

    # Use:
    #    AM.v = k = AC.v + R0 eR.v + p er.v (=0)
    # Re-arrange vs k dependence and multiply by R:
    #    R k = R AC.v + R R0 eR.v
    # Square:
    #    (R R0 eR.v)^2 = (R [k - AC.v])^2
    #                  = (R [k + CA.v])^2

    # Thus:
    # R0^2 [ RA eRA.v + k vpar^2 ]^2 = R^2 [ k + CA.v ]^2

    # R0^2 [ k^2 vpar^4 + 2 k vpar^2 RA (eRA.v) + RA^2 (eRA.v)^2 ]
    #    = [RA^2 + k^2 vpar^2 + 2 k RA (erA.v)] [ k^2 + 2k (CA.v) + (CA.v)^2 ]

    # Polynom deg 4 = 0
    # k^4:
    #   vpar^2
    # k^3:
    #   2(CA.v) vpar^2 + 2RA (eRA.v)
    # k^2:
    #   vpar^2 (CA.v)^2 + 4RA(erA.v)(CA.v) + RA^2 - R0^2 vpar^4
    # k:
    #   2RA (eRA.v)(CA.v)^2 + 2(CA.v)RA^2 - 2 R0^2 vpar^2 RA (eRA.v)
    # 0:
    #   RA^2 (CA.v)^2 - R0^2 RA^2 (eRA.v)^2

    CAv = ptsx * vectx + ptsy * vecty + (ptsz - Z0) * vectz
    vpar2 = vnorm2d**2
    RA = np.hypot(ptsx, ptsy)
    eRAv = np.cos(phi) * vectx + np.sin(phi) * vecty

    c4 = vpar2
    c3 = 2*CAv * vpar2 + 2*RA * eRAv
    c2 = vpar2 * CAv**2 + 4*RA*eRAv*CAv + RA**2 - R0**2 * vpar2**2
    c1 = 2*RA * eRAv * CAv**2 + 2*CAv*RA**2 - 2*R0**2 * vpar2 * RA * eRAv
    c0 = RA**2 * CAv**2 - R0**2 * RA**2 * eRAv**2

    # --------------------------------
    # loop on rays

    linds = [range(ss) for ss in ptsx.shape]

    dwarn = {}
    roots0 = np.full(ptsx.shape, np.nan)
    for ind in itt.product(*linds):

        coefs = np.r_[c0[ind], c1[ind], c2[ind], c3[ind], c4[ind]]
        if not np.all(np.isfinite(coefs)):
            continue

        # define polynomial
        poly = np.polynomial.polynomial.Polynomial(coefs)

        # roots
        roots = poly.roots()
        roots = np.real(roots[np.isreal(roots)])

        # remove solutions with R = 0
        if (ptsx[ind]**2 + ptsy[ind]**2 > 1.e-9) or vpar2[ind] > 1e-9:
            R2 = (
                RA[ind]**2
                + roots**2 * vpar2[ind]
                + 2 * roots * RA[ind] * eRAv[ind]
            )
            roots = roots[R2 > 1e-9]

        # positive only
        roots = roots[roots >= 0.]

        # only where scalar is close to 0
        Mx = ptsx[ind] + roots * vectx[ind]
        My = ptsy[ind] + roots * vecty[ind]
        Mz = ptsz[ind] + roots * vectz[ind]
        Mphi = np.arctan2(My, Mx)

        erx = Mx - R0*np.cos(Mphi)
        ery = My - R0*np.sin(Mphi)
        erz = Mz - Z0

        ern = np.sqrt(erx**2 + ery**2 + erz**2)
        erx = erx / ern
        ery = ery / ern
        erz = erz / ern

        # check dm.v = 0
        sca = (erx * vectx[ind] + ery * vecty[ind] + erz * vectz[ind])
        roots = roots[np.abs(sca) < 1e-9]

        # keep where Dphi < Dphimax
        Mphi = np.arctan2(
            ptsy[ind] + roots * vecty[ind],
            ptsx[ind] + roots * vectx[ind],
        )
        Dphi = np.abs(Mphi - phi[ind])
        discont = Dphi > np.pi
        Dphi[discont] = 2*np.pi - Dphi[discont]
        roots = roots[Dphi < Dphimax]

        # extract solution
        nsol = roots.size
        if nsol == 0:
            dwarn[ind] = "No >=0 roots found with R > 0 and Dphi < pi"
        else:
            if nsol > 1:
                dwarn[ind] = roots
            roots0[ind] = roots[np.argmin(np.abs(roots))]

    # ----------
    # warnings
    # ----------

    if len(dwarn) > 0 and verb == 2:
        _multiple_solutions(
            kray=kray,
            dwarn=dwarn,
            warn=True,
            # plot
            plot=True,
            R0=R0, Z0=Z0,
            ptsx=ptsx, ptsy=ptsy, ptsz=ptsz,
            vectx=vectx, vecty=vecty, vectz=vectz,
            length=length,
        )

    # -----------------------
    # initialize output
    # -----------------------

    impact = np.full(ptsx.shape, np.nan)
    ang = np.full(ptsx.shape, np.nan)

    rootsx = np.full(ptsx.shape, np.nan)
    rootsy = np.full(ptsx.shape, np.nan)
    rootsz = np.full(ptsx.shape, np.nan)

    # ---------------
    # store
    # ---------------

    iok = np.isfinite(roots0)

    rootsx[iok] = ptsx[iok] + roots0[iok] * vectx[iok]
    rootsy[iok] = ptsy[iok] + roots0[iok] * vecty[iok]
    rootsz[iok] = ptsz[iok] + roots0[iok] * vectz[iok]

    # --------------
    # safety check
    # --------------

    rootphi = np.arctan2(rootsy[iok], rootsx[iok])

    erx = rootsx[iok] - R0*np.cos(rootphi)
    ery = rootsy[iok] - R0*np.sin(rootphi)
    erz = rootsz[iok] - Z0

    ern = np.sqrt(erx**2 + ery**2 + erz**2)
    erx = erx / ern
    ery = ery / ern
    erz = erz / ern

    # check dm.v = 0
    sca = (erx * vectx[iok] + ery * vecty[iok] + erz * vectz[iok])
    iout = np.abs(sca) > 1e-9
    if np.any(iout):
        vnorm = np.sqrt(
            vectx[iok][iout]**2
            + vecty[iok][iout]**2
            + vectz[iok][iout]**2
        )
        ern = np.sqrt(erx**2 + ery**2 + erz**2)
        msg = (
            "Inconsistent sinogram solutions (non-perp to LOS)!\n"
            f"\t- rays: '{kray}'\n"
            f"\t- ind = {iout.nonzero()}\n"
            f"\t- sca = {sca[iout]}\n"
            f"\t- vectx = {vectx[iok][iout]}\n"
            f"\t- vecty = {vecty[iok][iout]}\n"
            f"\t- vectz = {vectz[iok][iout]}\n"
            f"\t- vect_norm = {vnorm}\n"
            f"\t- erx = {erx[iout]}\n"
            f"\t- ery = {ery[iout]}\n"
            f"\t- erz = {erz[iout]}\n"
            f"\t- er_norm = {ern}\n"
        )
        raise Exception(msg)

    # ---------------
    # impact and ang
    # ---------------

    impact[iok] = ern
    ang[iok] = np.arctan2(erz, erx*np.cos(rootphi) + ery*np.sin(rootphi))

    # -------------
    # format ouput
    # -------------

    dout = {
        'R0': R0,
        'Z0': Z0,
        'segment': segment,
        'root_kk': roots0,
        'root_ptx': rootsx,
        'root_pty': rootsy,
        'root_ptz': rootsz,
        'impact': impact,
        'ang': ang,
        'dphi': dphi,
    }

    return dout


# ###############################################################
# ###############################################################
#            Multiple solutions
# ###############################################################


def _multiple_solutions(
    kray=None,
    dwarn=None,
    warn=None,
    # plot
    plot=None,
    R0=None,
    Z0=None,
    ptsx=None, ptsy=None, ptsz=None,
    vectx=None, vecty=None, vectz=None,
    length=None,
):

    # ---------------------
    # pring warning
    # ---------------------

    if warn is True:
        lstr = [f"\t- {k0}: {v0}" for k0, v0 in dwarn.items()]
        msg = (
            "\nSinogram computation\n"
            f"The following rays of '{kray}' have non-unique solutions:\n"
            + "\n".join(lstr)
        )
        warnings.warn(msg)

    # ---------------------
    # plot
    # ---------------------

    dwarn_plot = {
        k0: v0 for k0, v0 in dwarn.items()
        if not isinstance(v0, str)
    }

    if plot is True and len(dwarn_plot) > 0:

        # ------------
        # prepare data

        theta = np.pi * np.linspace(-1, 1, 101)
        tit = f"Rays from '{kray}' with multiple sinogram roots"
        dax = {}

        # --------------
        # prepare figure

        fig = plt.figure(figsize=(12, 8))
        fig.suptitle(tit, size=14, fontweight='bold')

        ax0 = fig.add_axes([0.1, 0.1, 0.4, 0.8], projection='3d')
        ax0.set_xlabel('x (m)', size=12, fontweight='bold')
        ax0.set_ylabel('y (m)', size=12, fontweight='bold')
        ax0.set_zlabel('z (m)', size=12, fontweight='bold')

        dax['3d'] = ax0

        ax1 = fig.add_axes([0.5, 0.1, 0.4, 0.8], aspect='equal')
        ax1.set_xlabel('R (m)', size=12, fontweight='bold')
        ax1.set_ylabel('Z (m)', size=12, fontweight='bold')

        dax['cross'] = ax1

        # --------------
        # plot 3d

        # Reference
        ax0.plot(
            R0 * np.cos(theta),
            R0 * np.sin(theta),
            Z0 * np.ones(theta.shape),
            c='k',
            ls='-',
            marker='None',
        )

        # Lines
        for ind, roots in dwarn_plot.items():

            # LOS
            ax0.plot(
                ptsx[ind] + length[ind] * np.r_[0, 1] * vectx[ind],
                ptsy[ind] + length[ind] * np.r_[0, 1] * vecty[ind],
                ptsz[ind] + length[ind] * np.r_[0, 1] * vectz[ind],
                c='k',
                ls='-',
                lw=3.,
                marker='None',
                label=f'LOS_{ind}',
            )

            # mathematical line
            rootmax = np.max(roots)
            ax0.plot(
                ptsx[ind] + 1.5 * rootmax * np.r_[0, 1] * vectx[ind],
                ptsy[ind] + 1.5 * rootmax * np.r_[0, 1] * vecty[ind],
                ptsz[ind] + 1.5 * rootmax * np.r_[0, 1] * vectz[ind],
                c=(0.5, 0.5, 0.5),
                ls='-',
                lw=1.,
                marker='None',
                label=f'line_{ind}',
            )

            # roots
            rx = ptsx[ind] + roots * vectx[ind]
            ry = ptsy[ind] + roots * vecty[ind]
            rz = ptsz[ind] + roots * vectz[ind]

            rphi = np.arctan2(ry, rx)

            erx = R0 * np.cos(rphi)
            ery = R0 * np.sin(rphi)
            erz = Z0 * np.ones(rx.shape)

            nan = np.full(roots.shape, np.nan)
            px = np.array([rx, erx, nan]).T.ravel()
            py = np.array([ry, ery, nan]).T.ravel()
            pz = np.array([rz, erz, nan]).T.ravel()

            ax0.plot(
                px,
                py,
                pz,
                c='r',
                ls='-',
                lw=1.,
                label=f'perp_{ind}',
            )

        # ----------
        # plot cross

        ax1.plot(
            [R0],
            [Z0],
            marker='o',
            c='k',
        )

        # Lines
        for ind, roots in dwarn_plot.items():

            # LOS
            kk = np.linspace(0, length[ind], 100)
            R = np.hypot(
                ptsx[ind] + kk * vectx[ind],
                ptsy[ind] + kk * vecty[ind],
            )

            ax1.plot(
                R,
                ptsz[ind] + kk * vectz[ind],
                ls='-',
                lw=3,
                c='k',
                label=f'LOS_{ind}',
            )

            # max root
            rootmax = np.max(roots)
            kk = np.linspace(0, rootmax, 100)
            R = np.hypot(
                ptsx[ind] + kk * vectx[ind],
                ptsy[ind] + kk * vecty[ind],
            )

            ax1.plot(
                R,
                ptsz[ind] + kk * vectz[ind],
                ls='-',
                lw=1,
                c=(0.5, 0.5, 0.5),
                label=f'line_{ind}',
            )

            # roots
            rx = ptsx[ind] + roots * vectx[ind]
            ry = ptsy[ind] + roots * vecty[ind]
            rz = ptsz[ind] + roots * vectz[ind]

            rphi = np.arctan2(ry, rx)

            erx = R0 * np.cos(rphi)
            ery = R0 * np.sin(rphi)
            erz = Z0 * np.ones(rx.shape)

            nan = np.full(roots.shape, np.nan)
            px = np.array([rx, erx, nan]).T.ravel()
            py = np.array([ry, ery, nan]).T.ravel()
            pz = np.array([rz, erz, nan]).T.ravel()

            ax1.plot(
                np.hypot(px, py),
                pz,
                c='r',
                ls='-',
                lw=1.,
                label=f'perp_{ind}',
            )

        ax1.set_xlim(left=0)

    return


# ###############################################################
# ###############################################################
#               Compute - vessel
# ###############################################################


def _compute_config(
    config=None,
    kVes=None,
    R0=None,
    Z0=None,
    # options
    verb=None,
):

    # --------------
    # Select polygon

    polyR, polyZ = config.dStruct['dObj']['Ves'][kVes].dgeom['Poly']

    # ----------------
    # compute

    ang = np.pi * np.linspace(-1, 1, 101)

    pmax = np.max(
        (polyR[:, None] - R0) * np.cos(ang)[None, :]
        + (polyZ[:, None] - Z0) * np.sin(ang)[None, :],
        axis=0,
    )

    dout = {
        'ang': ang,
        'pmax': pmax,
    }

    return dout


# ###############################################################
# ###############################################################
#               Plot
# ###############################################################


def _plot(
    dout=None,
    dout_config=None,
    # sinogram ref point
    R0=None,
    Z0=None,
    # sinogram options
    ang=None,
    ang_units=None,
    impact_pos=None,
    pmax=None,
    # plotting options
    color=None,
    marker=None,
    label=None,
    sketch=None,
    # figure
    dax=None,
    dmargin=None,
    fs=None,
    wintit=None,
):

    # --------------
    # check

    (
        pmax,
        dprops,
        sketch,
        dax,
    ) = _check_plot(**locals())

    # -----------
    # prepare

    # -----------
    # plot rays

    # sinogram
    kax = 'sinogram'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        # plot rays
        for k0, v0 in dout.items():
            ax.plot(
                v0['ang'].ravel(),
                v0['impact'].ravel(),
                **dprops[k0],
            )

        # legend
        ax.legend(
            loc='upper left',
            bbox_to_anchor=(1, 1),
        )

        # -----------
        # plot config

        if dout_config is not None:

            inds = np.argsort(dout_config['ang'])
            if dout_config['impact_pos'] is True:
                ax.fill_between(
                    dout_config['ang'][inds],
                    dout_config['pmax'][inds],
                    pmax*np.ones((inds.size,)),
                    fc=(0.8, 0.8, 0.8),
                )

            else:
                ipos = dout_config['pmax'] >= 0
                inds = np.argsort(dout_config['ang'][ipos])
                ax.fill_between(
                    dout_config['ang'][ipos][inds],
                    dout_config['pmax'][ipos][inds],
                    pmax*np.ones((ipos.sum(),)),
                    fc=(0.8, 0.8, 0.8),
                )

                inds = np.argsort(dout_config['ang'][~ipos])
                ax.fill_between(
                    dout_config['ang'][~ipos][inds],
                    dout_config['pmax'][~ipos][inds],
                    -pmax*np.ones(((~ipos).sum(),)),
                    fc=(0.8, 0.8, 0.8),
                )

    # -----------
    # plot sketch

    kax = 'sketch'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        _plot_sketch(ax)

    return dax


# ###############################################################
# ###############################################################
#               Check plot
# ###############################################################


def _check_plot(
    dout=None,
    dout_config=None,
    # parameters
    R0=None,
    Z0=None,
    # angles
    ang=None,
    ang_units=None,
    impact_pos=None,
    pmax=None,
    # options
    color=None,
    marker=None,
    label=None,
    sketch=None,
    # figure
    dax=None,
    dmargin=None,
    fs=None,
    wintit=None,
    tit=None,
    # unused
    **kwdargs,
):

    # --------------
    # dprops

    dprops = {}
    for k0 in dout.keys():
        dprops[k0] = {
            'label': k0 if label is None else label,
            'color': color,
            'marker': '.' if marker is None else marker,
        }

    # -------------
    # sketch

    sketch = ds._generic_check._check_var(
        sketch, 'sketch',
        types=bool,
        default=True,
    )

    # -------------
    # pmax

    if pmax is None:

        lpmax = [np.nanmax(np.abs(v0['impact'])) for v0 in dout.values()]

        if dout_config is not None:
            lpmax.append(np.nanmax(np.abs(dout_config['pmax'])))

        if not np.any(np.isfinite(lpmax)):
            lstr = [f"\t- {k0}" for k0 in dout.keys()]
            msg = (
                "Impact parameters are all nans for:\n"
                + "\n".join(lstr)
                + f"\nlen(dout) = {len(dout)}\n"
                + f"dout_config is None: {dout_config is None}"
            )
            warnings.warn(msg)
            pmax = 1.

        else:
            pmax = np.nanmax(lpmax)

    pmax = float(ds._generic_check._check_var(
        pmax, 'pmax',
        types=(float, int),
        sign='>=0',
    ))

    # -------------
    # figure

    if fs is None:
        fs = (10, 8)

    if wintit is None:
        wintit = 'tofu - sinogram'

    if tit is None:
        tit = f"Sinogram with respect to (R, Z) = ({R0}, {Z0}) m"

    if dmargin is None:
        dmargin = {
            'left': 0.10, 'right': 0.95,
            'bottom': 0.10, 'top': 0.90,
            'wspace': 0.25, 'hspace': 0.1,
        }

    # -------------
    # sketch

    if dax is None:
        dax = _get_ax(
            ang=ang,
            ang_units=ang_units,
            impact_pos=impact_pos,
            pmax=pmax,
            # figure
            fs=fs,
            wintit=wintit,
            tit=tit,
            dmargin=dmargin,
        )

    else:
        if isinstance(dax, plt.Axes):
            dax = {'sinogram': {'handle': dax}}
        elif isinstance(dax, dict):
            for k0 in ['sinogram', 'sketch']:
                if isinstance(dax.get(k0), dict):
                    pass
                elif isinstance(dax.get(k0), plt.Axes):
                    dax[k0] = {'handle': dax[k0]}

    return (
        pmax,
        dprops,
        sketch,
        dax,
    )


def _get_ax(
    ang=None,
    ang_units=None,
    impact_pos=None,
    pmax=None,
    # figure
    dmargin=None,
    fs=None,
    wintit=None,
    tit=None,
):

    # create figure and axes
    fig = plt.figure(figsize=fs)
    fig.canvas.manager.set_window_title(wintit)
    fig.suptitle(tit, size=14, fontweight='bold')

    gs = GridSpec(ncols=6, nrows=6, **dmargin)

    # sinogram
    ax0 = fig.add_subplot(gs[:, :-1])
    xlab = r"$" + f"\{ang}" + r"$" + f" ({ang_units})"
    ax0.set_xlabel(xlab, size=12, fontweight='bold')
    ax0.set_ylabel(r"$p$ (m)", size=12, fontweight='bold')

    angmax = np.pi
    if ang_units == 'deg':
        angmax = 180

    if impact_pos is True:
        ax0.set_xlim(-angmax, angmax)
        ax0.set_ylim(0, pmax)
    else:
        if ang == 'theta':
            ax0.set_xlim(0, angmax)
        else:
            ax0.set_xlim(-angmax/2., angmax/2.)
        ax0.set_ylim(-pmax, pmax)

    # sketch
    ax1 = fig.add_subplot(
        gs[-1, -1],
        aspect='equal',
        adjustable='datalim',
        frameon=False,
    )
    ax1.set_xticks([])
    ax1.set_yticks([])

    # output dict
    dax = {
        'sinogram': {'handle': ax0},
        'sketch': {'handle': ax1},
    }

    return dax


# ##########################################################
# ##########################################################
#
# ##########################################################


def _plot_sketch(ax=None):

    # get points
    pt = np.array([[0, -0.8], [0, 0.8]])
    line = np.array([[-1.6, 0.1], [0, 1.7]])
    hor = np.array([[-0.4, 0.2], [1.2, 1.2]])
    vert = np.array([[-0.4, -0.4], [1.2, 1.7]])
    theta = np.linspace(0, 3.*np.pi/4., 30)
    ksi = np.linspace(0, np.pi/4., 10)
    zeta = np.linspace(np.pi/4., np.pi/2., 10)

    theta = np.array([0.3*np.cos(theta), 0.3*np.sin(theta)])
    ksi = np.array([-0.4+0.4*np.cos(ksi), 1.2+0.4*np.sin(ksi)])
    zeta = np.array([-0.4+0.4*np.cos(zeta), 1.2+0.4*np.sin(zeta)])

    # plot
    ax.plot(
        pt[0, :], pt[1, :], '+k',
        pt[0, :], pt[1, :], '--k',
        line[0, :], line[1, :], '-k',
        hor[0, :], hor[1, :], '-k',
        vert[0, :], vert[1, :], '-k',
        theta[0, :], theta[1, :], '-k',
        ksi[0, :], ksi[1, :], '-k',
        zeta[0, :], zeta[1, :], '-k',
    )

    # annotate
    ax.annotate(
        r"$\theta$",
        xy=(0.3, 0.4),
        xycoords='data',
        va="center",
        ha="center",
    )
    ax.annotate(
        r"$\xi$",
        xy=(0.1, 1.4),
        xycoords='data',
        va="center",
        ha="center",
    )
    ax.annotate(
        r"$\zeta$",
        xy=(-0.2, 1.7),
        xycoords='data',
        va="center",
        ha="center",
    )
    ax.annotate(
        r"$p$",
        xy=(-0.7, 0.3),
        xycoords='data',
        va="center",
        ha="center",
    )
    return



# -*- coding: utf-8 -*-


# Built-in
import warnings


# Common
import numpy as np
import datastock as ds


from . import _generic_plot
from . import _class09_compute_broadband as _compute_broadband


# #############################################################################
# #############################################################################
#                           Matrix - compute
# #############################################################################


def compute(
    coll=None,
    key=None,
    key_bsplines=None,
    key_diag=None,
    key_cam=None,
    # sampling
    res=None,
    mode=None,
    method=None,
    crop=None,
    dvos=None,
    # common ref
    ref_com=None,
    ref_vector_strategy=None,
    # options
    brightness=None,
    # output
    store=None,
    verb=None,
):
    """ Compute the geometry matrix using:
            - a Plasma2DRect instance with a key to a bspline set
            - a cam instance with a resolution
    """

    # -----------
    # check input
    # -----------

    (
        key,
        key_bs, key_m, ndim,
        subkey, key_bs0, key_m0,
        key_diag, key_cam,
        spectro,
        radius_max, method, res, mode, crop,
        dvos,
        brightness,
        store, verb,
    ) = _compute_check(
        coll=coll,
        key=key,
        key_bsplines=key_bsplines,
        key_diag=key_diag,
        key_cam=key_cam,
        # sampling
        method=method,
        res=res,
        mode=mode,
        crop=crop,
        dvos=dvos,
        # options
        brightness=brightness,
        # output
        store=store,
        verb=verb,
    )

    # -----------
    # prepare
    # -----------

    # prepare indices
    indbs = coll.select_ind(
        key=key_bs,
        returnas=bool,
        crop=crop,
    )

    # prepare slicing
    shape_mat, sli_mat, axis_pix, axis_bs, axis_other = _prepare(
        coll=coll,
        indbs=indbs,
        key_bs0=key_bs0,
        subkey=subkey,
    )

    if verb is True:
        msg = f"\nGeom matrix for diag '{key_diag}' and bs '{key_bs}':"
        print(msg)

    # -----------
    # compute
    # -----------

    if spectro is True:
        raise NotImplementedError("Spectro geom matrix not implemented!")

    else:

        if method == 'los':
            dout, axis = _compute_broadband._compute_los(
                coll=coll,
                key=key,
                key_bs=key_bs,
                key_diag=key_diag,
                key_cam=key_cam,
                # sampling
                indbs=indbs,
                res=res,
                mode=mode,
                radius_max=radius_max,
                # common ref
                ref_com=ref_com,
                ref_vector_strategy=ref_vector_strategy,
                # groupby=groupby,
                shape_mat=shape_mat,
                sli_mat=sli_mat,
                axis_pix=axis_pix,
                # other
                brightness=brightness,
                verb=verb,
            )

        else:
            dout, axis = _compute_broadband._compute_vos(
                coll=coll,
                key=key,
                key_bs=key_bs,
                key_diag=key_diag,
                key_cam=key_cam,
                # dvos
                dvos=dvos,
                # sampling
                indbs=indbs,
                res=res,
                mode=mode,
                radius_max=radius_max,
                # common ref
                ref_com=ref_com,
                ref_vector_strategy=ref_vector_strategy,
                # groupby=groupby,
                shape_mat=shape_mat,
                sli_mat=sli_mat,
                axis_pix=axis_pix,
                # other
                brightness=brightness,
                verb=verb,
            )

    # ---------------
    # check
    # ---------------

    if axis is None:
        _no_interaction(
            coll=coll,
            key=key,
            key_bs=key_bs,
            key_diag=key_diag,
            key_cam=key_cam,
        )
        store = False
        import pdb; pdb.set_trace() # DB

    # ---------------
    # store / return
    # ---------------

    if store:
        _store(
            coll=coll,
            key=key,
            key_bs=key_bs,
            key_diag=key_diag,
            key_cam=key_cam,
            method=method,
            res=res,
            crop=crop,
            dout=dout,
            axis_chan=axis,
            axis_bs=axis_bs,
            axis_other=axis_other,
        )

    else:
        return dout


# ###################
#   checking
# ###################


def _compute_check(
    coll=None,
    key=None,
    key_bsplines=None,
    key_diag=None,
    key_cam=None,
    # sampling
    method=None,
    res=None,
    mode=None,
    crop=None,
    dvos=None,
    # options
    brightness=None,
    # output
    store=None,
    verb=None,
):

    # --------------
    # keys

    wm = coll._which_mesh
    wbs = coll._which_bsplines

    # key
    key = ds._generic_check._obj_key(
        d0=coll.dobj.get('geom matrix', {}),
        short='gmat',
        key=key,
    )

    # key_diag, key_cam
    key_diag, key_cam = coll.get_diagnostic_cam(
        key=key_diag,
        key_cam=key_cam,
    )

    spectro = coll.dobj['diagnostic'][key_diag]['spectro']
    if spectro:
        msg = (
            "Geometry matrix can only be computed for non-spectro diags"
        )
        raise NotImplementedError(msg)

    # key_bs
    lk = list(coll.dobj.get(wbs, {}).keys())
    key_bs = ds._generic_check._check_var(
        key_bsplines, 'key_bsplines',
        types=str,
        allowed=lk,
    )

    # key_m
    key_m = coll.dobj[wbs][key_bs][wm]
    submesh = coll.dobj[wm][key_m]['submesh']
    if submesh is not None:
        key_m0 = submesh
        key_bs0 = coll.dobj[wm][key_m]['subbs']
        subkey = coll.dobj[wm][key_m]['subkey'][0]
    else:
        key_m0, key_bs0, subkey = None, None, None

    # -------------------
    # dimensions and axis

    if submesh is None:
        key_kR = coll.dobj[wm][key_m]['knots'][0]
        ndim = len(coll.dobj[wbs][key_bs]['shape'])
    else:
        key_kR = coll.dobj[wm][key_m0]['knots'][0]
        ndim = len(coll.ddata[subkey]['shape'])

    radius_max = np.max(coll.ddata[key_kR]['data'])

    # --------------
    # parameters

    # method
    method = ds._generic_check._check_var(
        method, 'method',
        default='los',
        types=str,
        allowed=['los', 'vos'],
    )

    # res
    res = ds._generic_check._check_var(
        res, 'res',
        default=0.01,
        types=float,
        sign='> 0.',
    )

    # mode
    mode = ds._generic_check._check_var(
        mode, 'mode',
        default='abs',
        types=str,
        allowed=['abs', 'rel'],
    )

    # crop
    crop = ds._generic_check._check_var(
        crop, 'crop',
        default=True,
        types=bool,
    )
    crop = (
        crop
        and coll.dobj[wbs][key_bs]['crop'] not in [None, False]
    )

    # dvos
    if method == 'vos':
        key_diag, dvos, isstore = coll.check_diagnostic_dvos(
            key=key_diag,
            key_cam=key_cam,
            dvos=dvos,
        )

    # brightness
    brightness = ds._generic_check._check_var(
        brightness, 'brightness',
        types=bool,
        default=False,
    )

    # store
    store = ds._generic_check._check_var(
        store, 'store',
        default=True,
        types=bool,
    )

    # verb
    if verb is None:
        verb = True
    if not isinstance(verb, bool):
        msg = (
            f"Arg verb must be a bool!\n"
            f"\t- provided: {verb}"
        )
        raise Exception(msg)

    return (
        key,
        key_bs, key_m, ndim,
        subkey, key_bs0, key_m0,
        key_diag, key_cam,
        spectro,
        radius_max, method, res, mode, crop,
        dvos,
        brightness,
        store, verb,
    )


# ###################
#   prepare
# ###################


def _prepare(
    coll=None,
    indbs=None,
    key_bs0=None,
    subkey=None,
):

    # shapes
    nbs = indbs.sum()
    nchan = None        # depends on cam

    # cases
    wbs = coll._which_bsplines
    if subkey is None:
        shape_mat = (nchan, nbs)

        sli_mat = [None, slice(None)]
        axis_pix = 0
        axis_other = None

    else:

        refbs = coll.dobj[wbs][key_bs0]['ref']
        ref = coll.ddata[subkey]['ref']
        sh = list(coll.ddata[subkey]['shape'])
        axis_pix = ref.index(refbs[0])

        if len(refbs) == 1:  # tri
            sh = sh[:axis_pix+1] + [None] + sh[axis_pix+1:]
        axis_bs = axis_pix + 1

        axis_other = [ii for ii, rr in enumerate(ref) if rr not in refbs]
        if len(axis_other) == 1:
            axis_other = axis_other[0]
            sh[axis_pix] = nchan
            sh[axis_bs] = nbs
            shape_mat = tuple(sh)

            sli_mat = [None, None, None]
            sli_mat[axis_bs] = slice(None)
            sli_mat[axis_other] = slice(None)
        else:
            shape_mat = (nchan, nbs)

            sli_mat = [None, slice(None)]
            axis_pix = 0
            axis_other = None

    axis_bs = axis_pix + 1

    return shape_mat, sli_mat, axis_pix, axis_bs, axis_other


# ###################
#   storing
# ###################


def _store(
    coll=None,
    key=None,
    key_bs=None,
    key_diag=None,
    key_cam=None,
    method=None,
    res=None,
    crop=None,
    dout=None,
    axis_chan=None,
    axis_bs=None,
    axis_other=None,
):

    # shapes
    shapes = [v0['data'].shape for v0 in dout.values()]
    assert all([len(ss) == len(shapes[0]) for ss in shapes[1:]])
    shapes = np.array(shapes)
    assert np.allclose(shapes[1:, :axis_chan], shapes[0:1, :axis_chan])
    assert np.allclose(shapes[1:, axis_bs:], shapes[0:1, axis_bs:])

    # add matrix obj
    dobj = {
        'geom matrix': {
            key: {
                'data': list(dout.keys()),
                'bsplines': key_bs,
                'diagnostic': key_diag,
                'camera': key_cam,
                'method': method,
                'res': res,
                'crop': crop,
                'shape': tuple(shapes[0, :]),
                'axis_chan': axis_chan,
                'axis_bs': axis_bs,
                'axis_other': axis_other,
            },
        },
    }

    coll.update(ddata=dout, dobj=dobj)


# ##################################################################
# ##################################################################
#               retrofit
# ##################################################################


def _concatenate(
    coll=None,
    key=None,
    key_cam=None,
):

    # ------------
    # check inputs

    lok = list(coll.dobj.get('geom matrix', {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    key_cam0 = coll.dobj['geom matrix'][key]['camera']
    key_cam = ds._generic_check._check_var_iter(
        key_cam, 'key_cam',
        default=key_cam0,
        types=(list, tuple),
        types_iter=str,
        allowed=key_cam0,
    )

    # -----------
    # concatenate

    key_data = coll.dobj['geom matrix'][key]['data']
    axis = coll.dobj['geom matrix'][key]['axis_chan']

    ref = list(coll.ddata[key_data[0]]['ref'])
    ref[axis] = None

    ind = 0
    dind = {}
    ldata = []
    for ii, k0 in enumerate(key_cam):
        datai = coll.ddata[key_data[ii]]['data']
        dind[k0] = ind + np.arange(0, datai.shape[axis])
        ldata.append(datai)
        ind += datai.shape[axis]

    data = np.concatenate(ldata, axis=axis)

    return data, ref, dind


# ###################
#   no interaction
# ###################


def _no_interaction(
    coll=None,
    key=None,
    key_bs=None,
    key_diag=None,
    key_cam=None,
):

    # ----------
    # plot

    wm = coll._which_mesh
    wbs = coll._which_bsplines
    keym = coll.dobj[wbs][key_bs][wm]
    submesh = coll.dobj[wm][keym]['submesh']

    is2d = coll.dobj['diagnostic'][key_diag]['is2d']

    # prepare dax
    dax0 = _generic_plot.get_dax_diag(
        proj=['cross', 'hor', '3d', 'camera'],
        dmargin=None,
        fs=None,
        wintit=None,
        tit='debug',
        is2d=is2d,
        key_cam=key_cam,
    )

    # mesh
    if submesh is None:
        _ = coll.plot_mesh(
            key=keym,
            dax={'cross': dax0['cross']},
            crop=True,
        )

    else:
        _ = coll.plot_mesh(
            key=submesh,
            dax={'cross': dax0['cross']},
            crop=True,
        )

        _ = coll.plot_mesh(keym)

    # cam
    _ = coll.plot_diagnostic(
        key=key_diag,
        key_cam=key_cam,
        elements='o',
        dax=dax0,
    )

    # -----
    # msg

    msg = (
        "No interaction detected between:\n"
        f"\t- camera: '{key_cam}'\n"
        f"\t- bsplines: '{key_bs}'\n"
        f"\t- submesh: '{submesh}'\n"
    )
    warnings.warn(msg)


# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: cdivision=True
#
################################################################################
# Utility functions related to the distance between a LOS and a VPoly:
#    - Compute distance between LOS and a circle (unit and vectorial)
#    - Test if a LOS and a circle are close (unit and vectorial) to a epsilon
#    - Compute distance between LOS and VPoly (unit and vectorial)
#    - Test if a LOS and a VPoly are close (unit and vectorial) to a epsilon
#    - Compute which LOS is closer to a VPoly
#    - Compute which VPoly is closer to a LOS
################################################################################
cimport cython
from cython.parallel import prange
from cython.parallel cimport parallel
from libc.math cimport fabs as c_abs
from libc.math cimport sqrt as c_sqrt
from libc.math cimport NAN as C_NAN
from libc.stdlib cimport malloc, free
# from libc.stdio cimport printf    # for debug
from ._basic_geom_tools cimport _VSMALL
from . cimport _basic_geom_tools as _bgt
from . cimport _sampling_tools as _st

# ==============================================================================
# == DISTANCE CIRCLE - LOS
# ==============================================================================
cdef inline void dist_los_circle_core(const double[3] direct,
                                      const double[3] origin,
                                      const double radius, const double circ_z,
                                      double norm_dir,
                                      double[2] result) nogil:
    """
    This function computes the intersection of a Ray (or Line Of Sight)
    and a horizontal circle in 3. It returns `kmin` the coefficient such that
    the ray of origin O = [ori1, ori2, ori3] and of directional vector
    D = [dir1, dir2, dir3] is closest to the circle of radius `radius`,
    center `(0, 0, circ_z)` and of normal (0,0,1) at the point P = O + kmin * D.
    And `distance` the distance between the two closest points
    The variable `norm_dir` is the norm of the direction of the ray.
    if you haven't normalized the ray (and for optimization reasons you dont
    want to, you can pass norm_dir = -1
    ---
    Source: https://www.geometrictools.com/Documentation/DistanceToCircle3.pdf
    The line is P(t) = B+t*M.  The circle is |X-C| = r with Dot(N,X-C)=0.

    Params
    ======
    direct : double (3) array
       directional vector of the ray
    origin : double (3) array
       origin of the array (in 3d)
    radius : double
       radius of the circle
    circ_z : double
       3rd coordinate of the center of the circle
       ie. the circle center is (0,0, circ_z)
    norm_dir : double (3) array
       normal of the direction of the vector (for computation performance)
    result : double (2) array
       - result[0] will contain the k coefficient to find the line point closest
       closest point
       - result[1] will contain the DISTANCE from line closest point to circle
       to the circle
    """
    cdef int numRoots, i
    cdef double zero = 0., m0sqr, m0, rm0
    cdef double lambd, m2b2, b1sqr, b1, twoThirds, sHat, gHat, cutoff, s
    cdef double[3] D
    cdef double[3] MxN
    cdef double[3] DxN
    cdef double[3] NxDelta
    cdef double[3] circle_normal
    cdef double[3] roots
    cdef double[3] diff
    cdef double[3] direction
    cdef double[3] line_closest
    cdef double[3] circle_center
    cdef double[3] circle_closest
    cdef double tmin
    cdef double distance
    cdef double inv_norm_dir

    if norm_dir < 0:
        norm_dir = c_sqrt(_bgt.compute_dot_prod(direct, direct))
    inv_norm_dir = 1./ norm_dir
    # .. initialization .....
    for i in range(3):
        circle_center[i] = 0.
        circle_normal[i] = 0.
        roots[i] = 0.
        # we normalize direction
        direction[i] = direct[i] * inv_norm_dir
    circle_normal[2] = 1
    circle_center[2] = circ_z

    D[0] = origin[0]
    D[1] = origin[1]
    D[2] = origin[2] - circ_z
    _bgt.compute_cross_prod(direction, circle_normal, MxN)
    _bgt.compute_cross_prod(D, circle_normal, DxN)
    m0sqr = _bgt.compute_dot_prod(MxN, MxN)

    if (m0sqr > zero):
        # Compute the critical points s for F'(s) = 0.
        numRoots = 0

        # The line direction M and the plane normal N are not parallel.  Move
        # the line origin B = (b0,b1,b2) to B' = B + lambd*direction =
        # (0,b1',b2').
        m0 = c_sqrt(m0sqr)
        rm0 = radius * m0
        lambd = -_bgt.compute_dot_prod(MxN, DxN) / m0sqr
        for i in range(3):
            D[i] += lambd * direction[i]
            DxN[i] += lambd * MxN[i]
        m2b2 = _bgt.compute_dot_prod(direction, D)
        b1sqr = _bgt.compute_dot_prod(DxN, DxN)
        if (b1sqr > zero) :
            # B' = (0,b1',b2') where b1' != 0.  See Sections 1.1.2 and 1.2.2
            # of the PDF documentation.
            b1 = c_sqrt(b1sqr)
            rm0sqr = radius * m0sqr
            if (rm0sqr > b1):
                twoThirds = 2.0 / 3.0
                sHat = c_sqrt(c_abs((rm0sqr * b1sqr)**twoThirds - b1sqr)) / m0
                gHat = rm0sqr * sHat / c_sqrt(c_abs(m0sqr * sHat * sHat + b1sqr))
                cutoff = gHat - sHat
                if (m2b2 <= -cutoff):
                    s = _bgt.compute_bisect(m2b2, rm0sqr, m0sqr, b1sqr, -m2b2, -m2b2 + rm0)
                    roots[numRoots] = s
                    numRoots += 1
                    if (m2b2 == -cutoff):
                        roots[numRoots] = -sHat
                        numRoots += 1
                elif (m2b2 >= cutoff):
                    s = _bgt.compute_bisect(m2b2, rm0sqr, m0sqr, b1sqr, -m2b2 - rm0,
                        -m2b2)
                    roots[numRoots] = s
                    numRoots += 1
                    if (m2b2 == cutoff):
                        roots[numRoots] = sHat
                        numRoots += 1
                else:
                    if (m2b2 <= zero):
                        s = _bgt.compute_bisect(m2b2, rm0sqr, m0sqr, b1sqr, -m2b2,
                            -m2b2 + rm0)
                        roots[numRoots] = s
                        numRoots += 1
                        s = _bgt.compute_bisect(m2b2, rm0sqr, m0sqr, b1sqr, -m2b2 - rm0,
                            -sHat)
                        roots[numRoots] = s
                        numRoots += 1
                    else:
                        s = _bgt.compute_bisect(m2b2, rm0sqr, m0sqr, b1sqr, -m2b2 - rm0,
                            -m2b2)
                        roots[numRoots] = s
                        numRoots += 1
                        s = _bgt.compute_bisect(m2b2, rm0sqr, m0sqr, b1sqr, sHat,
                            -m2b2 + rm0)
                        roots[numRoots] = s
                        numRoots += 1
            else:
                if (m2b2 < zero):
                    s = _bgt.compute_bisect(m2b2, rm0sqr, m0sqr, b1sqr, -m2b2,
                        -m2b2 + rm0)
                elif (m2b2 > zero):
                    s = _bgt.compute_bisect(m2b2, rm0sqr, m0sqr, b1sqr, -m2b2 - rm0,
                        -m2b2)
                else:
                    s = zero
                roots[numRoots] = s
                numRoots += 1
        else:
            # The new line origin is B' = (0,0,b2').
            if (m2b2 < zero):
                s = -m2b2 + rm0
                roots[numRoots] = s
                numRoots += 1
            elif (m2b2 > zero):
                s = -m2b2 - rm0
                roots[numRoots] = s
                numRoots += 1
            else:
                s = -m2b2 + rm0
                roots[numRoots] = s
                numRoots += 1
                s = -m2b2 - rm0
                roots[numRoots] = s
                numRoots += 1
        # Checking which one is the closest solution............................
        tmin = roots[0] + lambd
        for i in range(1,numRoots):
            t = roots[i] + lambd
            if (t>0 and t<tmin):
                tmin = t
        if tmin < 0:
            tmin = 0.
        # Now that we know the closest point on the line we can compute the
        # closest point on the circle and compute the distance
        line_closest[0] = origin[0] + tmin * direction[0]
        line_closest[1] = origin[1] + tmin * direction[1]
        line_closest[2] = origin[2] + tmin * direction[2]
        _bgt.compute_cross_prod(circle_normal, line_closest, NxDelta)
        if not (c_abs(NxDelta[0]) <= _VSMALL
                and c_abs(NxDelta[1]) <= _VSMALL
                and c_abs(NxDelta[2]) <= _VSMALL):
            norm_ppar = c_sqrt(line_closest[0]*line_closest[0]
                              + line_closest[1]*line_closest[1])
            circle_closest[0] = radius * line_closest[0] / norm_ppar
            circle_closest[1] = radius * line_closest[1] / norm_ppar
            circle_closest[2] = circle_center[2]
            for i in range(3):
                diff[i] = line_closest[i] - circle_closest[i]
            distance = c_sqrt(_bgt.compute_dot_prod(diff, diff))
            result[0] = tmin
            result[1] = distance
        else:
            diff[0] = line_closest[0] - radius
            diff[1] = line_closest[1]
            diff[2] = line_closest[2] - circle_center[2]
            distance = c_sqrt(_bgt.compute_dot_prod(diff, diff))
            result[0] = tmin
            result[1] = distance
    else:
        # The line direction and the plane normal are parallel.
        # There is only one solution the intersection between line and plane
        if not (c_abs(DxN[0]) <= _VSMALL
                and c_abs(DxN[1]) <= _VSMALL
                and c_abs(DxN[2]) <= _VSMALL):
            # The line is A+t*N but with A != C.
            t = -_bgt.compute_dot_prod(direction, D)
            # We compute line closest
            line_closest[0] = origin[0] + t * direction[0]
            line_closest[1] = origin[1] + t * direction[1]
            line_closest[2] = origin[2] + t * direction[2]
            # We compute cirlce closest
            for i in range(3):
                diff[i] = line_closest[i] - circle_center[i]
            distance = radius / c_sqrt(_bgt.compute_dot_prod(diff, diff))
            circle_closest[0] = line_closest[0] * distance
            circle_closest[1] = line_closest[1] * distance
            circle_closest[2] = circ_z + (line_closest[2] - circ_z) * distance
            if t < 0:
                # fi t is negative, we take origin as closest point
                t = 0.
                line_closest[0] = origin[0]
                line_closest[1] = origin[1]
                line_closest[2] = origin[2]
            for i in range(3):
                diff[i] = line_closest[i] - circle_closest[i]
            distance = c_sqrt(_bgt.compute_dot_prod(diff, diff))
            result[0] = t
            result[1] = distance
        else:
            # The line direction and the normal vector are on the same line
            # so C is the closest point for the circle and the distance is
            # the radius unless the ray's origin is after the circle center
            if (origin[2] * direction[2] <= circle_center[2] * direction[2]) :
                t = c_abs(circle_center[2] - origin[2])
                result[0] = t
                result[1] = radius
            else:
                t = c_abs(circle_center[2] - origin[2])
                result[0] = 0
                result[1] = c_sqrt(radius*radius + t*t)
    result[0] = result[0] * inv_norm_dir
    return

cdef inline void comp_dist_los_circle_vec_core(int num_los, int num_cir,
                                               double* los_directions,
                                               double* los_origins,
                                               double* circle_radius,
                                               double* circle_z,
                                               double* norm_dir_tab,
                                               double[::1] res_k,
                                               double[::1] res_dist,
                                               int num_threads) nogil:
    """ This function computes the intersection of a Ray (or Line Of Sight)
    # and a circle in 3D. It returns `kmin`, the coefficient such that the
    # ray of origin O = [ori1, ori2, ori3] and of directional vector
    # D = [dir1, dir2, dir3] is closest to the circle of radius `radius`
    # and centered `(0, 0, circ_z)` at the point P = O + kmin * D.
    # The variable `norm_dir` is the squared norm of the direction of the ray.
    # This is the vectorial version, we expect the directions and origins to be:
    # dirs = [dir1_los1, dir2_los1, dir3_los1, dir1_los2,...]
    # oris = [ori1_los1, ori2_los1, ori3_los1, ori1_los2,...]
    # res = [kmin(los1, cir1), kmin(los1, cir2),...]
    # ---
    # This is the PYTHON function, use only if you need this computation from
    # Python, if you need it from cython, use `dist_los_circle_core`
    """
    cdef int i, ind_los, ind_cir
    cdef double* loc_res
    cdef double* dirv
    cdef double* orig
    cdef double radius, circ_z, norm_dir
    with nogil, parallel(num_threads=num_threads):
        dirv = <double*>malloc(3*sizeof(double))
        orig = <double*>malloc(3*sizeof(double))
        loc_res = <double*>malloc(2*sizeof(double))
        for ind_los in prange(num_los):
            for i in range(3):
                dirv[i] = los_directions[ind_los * 3 + i]
                orig[i] = los_origins[ind_los * 3 + i]
            norm_dir = norm_dir_tab[ind_los]
            if norm_dir < 0.:
                norm_dir = c_sqrt(_bgt.compute_dot_prod(dirv, dirv))
            for ind_cir in range(num_cir):
                radius = circle_radius[ind_cir]
                circ_z = circle_z[ind_cir]
                dist_los_circle_core(dirv, orig, radius, circ_z,
                                     norm_dir, loc_res)
                res_k[ind_los * num_cir + ind_cir] = loc_res[0]
                res_dist[ind_los * num_cir + ind_cir] = loc_res[1]
        free(dirv)
        free(orig)
        free(loc_res)
    return

# ==============================================================================
# == TEST CLOSENESS CIRCLE - LOS
# ==============================================================================
cdef inline bint is_close_los_circle_core(const double[3] direct,
                                          const double[3] origin,
                                          double radius, double circ_z,
                                          double norm_dir, double eps) nogil:
    # Source: https://www.geometrictools.com/Documentation/DistanceToCircle3.pdf
    # The line is P(t) = B+t*M.  The circle is |X-C| = r with Dot(N,X-C)=0.
    cdef int numRoots, i
    cdef double zero = 0., m0sqr, m0, rm0
    cdef double lambd, m2b2, b1sqr, b1, twoThirds, sHat, gHat, cutoff, s
    cdef double[3] D
    cdef double[3] MxN
    cdef double[3] DxN
    cdef double[3] NxDelta
    cdef double[3] circle_normal
    cdef double[3] roots
    cdef double[3] diff
    cdef double[3] circle_center
    cdef double[3] circle_closest
    cdef double[3] line_closest
    cdef double[3] direction
    cdef double tmin
    cdef double distance
    cdef double inv_norm_dir
    cdef bint are_close

    # .. initialization .....
    if norm_dir < 0:
        norm_dir = c_sqrt(_bgt.compute_dot_prod(direct, direct))
    inv_norm_dir = 1./ norm_dir
    # .. initialization .....
    for i in range(3):
        circle_center[i] = 0.
        circle_normal[i] = 0.
        roots[i] = 0.
        # we normalize direction
        direction[i] = direct[i] * inv_norm_dir
    circle_normal[2] = 1
    circle_center[2] = circ_z

    D[0] = origin[0]
    D[1] = origin[1]
    D[2] = origin[2] - circ_z
    _bgt.compute_cross_prod(direction, circle_normal, MxN)
    _bgt.compute_cross_prod(D, circle_normal, DxN)
    m0sqr = _bgt.compute_dot_prod(MxN, MxN)

    if (m0sqr > zero):
        # Compute the critical points s for F'(s) = 0.
        numRoots = 0
        # The line direction M and the plane normal N are not parallel.  Move
        # the line origin B = (b0,b1,b2) to B' = B + lambd*direction =
        # (0,b1',b2').
        m0 = c_sqrt(m0sqr)
        rm0 = radius * m0
        lambd = -_bgt.compute_dot_prod(MxN, DxN) / m0sqr
        for i in range(3):
            D[i] += lambd * direction[i]
            DxN[i] += lambd * MxN[i]
        m2b2 = _bgt.compute_dot_prod(direction, D)
        b1sqr = _bgt.compute_dot_prod(DxN, DxN)
        if (b1sqr > zero) :
            # B' = (0,b1',b2') where b1' != 0.  See Sections 1.1.2 and 1.2.2
            # of the PDF documentation.
            b1 = c_sqrt(b1sqr)
            rm0sqr = radius * m0sqr
            if (rm0sqr > b1):
                twoThirds = 2.0 / 3.0
                sHat = c_sqrt(c_abs((rm0sqr * b1sqr)**twoThirds - b1sqr)) / m0
                gHat = rm0sqr * sHat / c_sqrt(c_abs(m0sqr * sHat * sHat + b1sqr))
                cutoff = gHat - sHat
                if (m2b2 <= -cutoff):
                    s = _bgt.compute_bisect(m2b2, rm0sqr, m0sqr, b1sqr, -m2b2, -m2b2 + rm0)
                    roots[numRoots] = s
                    numRoots += 1
                    if (m2b2 == -cutoff):
                        roots[numRoots] = -sHat
                        numRoots += 1
                elif (m2b2 >= cutoff):
                    s = _bgt.compute_bisect(m2b2, rm0sqr, m0sqr, b1sqr, -m2b2 - rm0,
                        -m2b2)
                    roots[numRoots] = s
                    numRoots += 1
                    if (m2b2 == cutoff):
                        roots[numRoots] = sHat
                        numRoots += 1
                else:
                    if (m2b2 <= zero):
                        s = _bgt.compute_bisect(m2b2, rm0sqr, m0sqr, b1sqr, -m2b2,
                            -m2b2 + rm0)
                        roots[numRoots] = s
                        numRoots += 1
                        s = _bgt.compute_bisect(m2b2, rm0sqr, m0sqr, b1sqr, -m2b2 - rm0,
                            -sHat)
                        roots[numRoots] = s
                        numRoots += 1
                    else:
                        s = _bgt.compute_bisect(m2b2, rm0sqr, m0sqr, b1sqr, -m2b2 - rm0,
                            -m2b2)
                        roots[numRoots] = s
                        numRoots += 1
                        s = _bgt.compute_bisect(m2b2, rm0sqr, m0sqr, b1sqr, sHat,
                            -m2b2 + rm0)
                        roots[numRoots] = s
                        numRoots += 1
            else:
                if (m2b2 < zero):
                    s = _bgt.compute_bisect(m2b2, rm0sqr, m0sqr, b1sqr, -m2b2,
                        -m2b2 + rm0)
                elif (m2b2 > zero):
                    s = _bgt.compute_bisect(m2b2, rm0sqr, m0sqr, b1sqr, -m2b2 - rm0,
                        -m2b2)
                else:
                    s = zero
                roots[numRoots] = s
                numRoots += 1
        else:
            # The new line origin is B' = (0,0,b2').
            if (m2b2 < zero):
                s = -m2b2 + rm0
                roots[numRoots] = s
                numRoots += 1
            elif (m2b2 > zero):
                s = -m2b2 - rm0
                roots[numRoots] = s
                numRoots += 1
            else:
                s = -m2b2 + rm0
                roots[numRoots] = s
                numRoots += 1
                s = -m2b2 - rm0
                roots[numRoots] = s
                numRoots += 1
        # Checking which one is the closest solution............................
        tmin = roots[0] + lambd
        for i in range(1,numRoots):
            t = roots[i] + lambd
            if (t>0 and t<tmin):
                tmin = t
        if tmin < 0:
            tmin = 0.
        # Now that we know the closest point on the line we can compute the
        # closest point on the circle and compute the distance
        line_closest[0] = origin[0] + tmin * direction[0]
        line_closest[1] = origin[1] + tmin * direction[1]
        line_closest[2] = origin[2] + tmin * direction[2]
        _bgt.compute_cross_prod(circle_normal, line_closest, NxDelta)
        if not (c_abs(NxDelta[0]) <= _VSMALL
                and c_abs(NxDelta[1]) <= _VSMALL
                and c_abs(NxDelta[2]) <= _VSMALL):
            norm_ppar = c_sqrt(line_closest[0]*line_closest[0]
                              + line_closest[1]*line_closest[1])
            circle_closest[0] = radius * line_closest[0] / norm_ppar
            circle_closest[1] = radius * line_closest[1] / norm_ppar
            circle_closest[2] = circle_center[2]
            for i in range(3):
                diff[i] = line_closest[i] - circle_closest[i]
            distance = c_sqrt(_bgt.compute_dot_prod(diff, diff))
            are_close = distance < eps
            return are_close
        else:
            diff[0] = line_closest[0] + radius
            diff[1] = line_closest[1]
            diff[2] = line_closest[2] - circle_center[2]
            distance = c_sqrt(_bgt.compute_dot_prod(diff, diff))
            are_close = distance < eps
            return are_close
    else:
        # The line direction and the plane normal are parallel.
        # There is only one solution the intersection between line and plane
        if not (c_abs(DxN[0]) <= _VSMALL
                and c_abs(DxN[1]) <= _VSMALL
                and c_abs(DxN[2]) <= _VSMALL):
            # The line is A+t*N but with A != C.
            t = -_bgt.compute_dot_prod(direction, D)
            # We compute line closest
            line_closest[0] = origin[0] + t * direction[0]
            line_closest[1] = origin[1] + t * direction[1]
            line_closest[2] = origin[2] + t * direction[2]
            # We compute cirlce closest
            for i in range(3):
                diff[i] = line_closest[i] - circle_center[i]
            distance = radius / c_sqrt(_bgt.compute_dot_prod(diff, diff))
            circle_closest[0] = line_closest[0] * distance
            circle_closest[1] = line_closest[1] * distance
            circle_closest[2] = circ_z + (line_closest[2] - circ_z) * distance
            if t < 0:
                # fi t is negative, we take origin as closest point
                line_closest[0] = origin[0]
                line_closest[1] = origin[1]
                line_closest[2] = origin[2]
            for i in range(3):
                diff[i] = line_closest[i] - circle_closest[i]
            distance = c_sqrt(_bgt.compute_dot_prod(diff, diff))
            are_close = distance < eps
            return are_close
        else:
            # The line direction and the normal vector are on the same line
            # so C is the closest point for the circle and the distance is
            # the radius unless the ray's origin is after the circle center
            if (origin[2] * direction[2] <= circle_center[2] * direction[2]) :
                are_close = radius < eps
                return are_close
            else:
                t = c_abs(circle_center[2] - origin[2])
                are_close = c_sqrt(radius*radius + t*t) < eps
                return are_close

cdef inline void is_close_los_circle_vec_core(int num_los, int num_cir,
                                              double eps,
                                              double* los_directions,
                                              double* los_origins,
                                              double* circle_radius,
                                              double* circle_z,
                                              double* norm_dir_tab,
                                              int[::1] res,
                                              int num_threads) nogil:
    """
    This function computes the intersection of a Ray (or Line Of Sight)
    and a circle in 3D. It returns `kmin`, the coefficient such that the
    ray of origin O = [ori1, ori2, ori3] and of directional vector
    D = [dir1, dir2, dir3] is closest to the circle of radius `radius`
    and centered `(0, 0, circ_z)` at the point P = O + kmin * D.
    The variable `norm_dir` is the squared norm of the direction of the ray.
    This is the vectorial version, we expect the directions and origins to be:
    dirs = [dir1_los1, dir2_los1, dir3_los1, dir1_los2,...]
    oris = [ori1_los1, ori2_los1, ori3_los1, ori1_los2,...]
    res = [kmin(los1, cir1), kmin(los1, cir2),...]
    ---
    This is the PYTHON function, use only if you need this computation from
    Python, if you need it from cython, use `dist_los_circle_core`
    """
    cdef int i, ind_los, ind_cir
    cdef double* dirv
    cdef double* orig
    cdef double radius, circ_z, norm_dir
    with nogil, parallel(num_threads=num_threads):
        dirv = <double*>malloc(3*sizeof(double))
        orig = <double*>malloc(3*sizeof(double))
        for ind_los in prange(num_los):
            for i in range(3):
                dirv[i] = los_directions[ind_los * 3 + i]
                orig[i] = los_origins[ind_los * 3 + i]
            norm_dir = norm_dir_tab[ind_los]
            if norm_dir < 0.:
                norm_dir = c_sqrt(_bgt.compute_dot_prod(dirv, dirv))
            for ind_cir in range(num_cir):
                radius = circle_radius[ind_cir]
                circ_z = circle_z[ind_cir]
                res[ind_los * num_cir
                    + ind_cir] = is_close_los_circle_core(dirv, orig, radius,
                                                          circ_z, norm_dir, eps)
        free(dirv)
        free(orig)
    return

# ==============================================================================
# == DISTANCE BETWEEN LOS AND EXT-POLY
# ==============================================================================
cdef inline void comp_dist_los_vpoly_vec_core(int num_poly, int nlos,
                                              double* ray_orig,
                                              double* ray_vdir,
                                              double[:,:,::1] ves_poly,
                                              double eps_uz,
                                              double eps_a,
                                              double eps_vz,
                                              double eps_b,
                                              double eps_plane,
                                              int ves_type,
                                              int algo_type,
                                              double* res_k,
                                              double* res_dist,
                                              double disc_step,
                                              int num_threads) nogil:
    """
    This function computes the distance (and the associated k) between nlos
    Rays (or LOS) and several `IN` structures (polygons extruded around the axis
    (0,0,1), eg. flux surfaces).
    For more details on the algorithm please see PDF: <name_of_pdf>.pdf #TODO

    Params
    ======
        num_poly : int
           Number of flux surfaces
        nlos : int
           Number of LOS
        ray_orig : (3, nlos) double array
           LOS origin points coordinates
        ray_vdir : (3, nlos) double array
           LOS normalized direction vector
        ves_poly : (num_pol, 2, num_vertex) double array
           Coordinates of the vertices of the Polygon defining the 2D poloidal
           cut of the different IN surfaces.
           WARNING : we suppose all poly are nested in each other,
                     from inner to outer
        eps_<val> : double
           Small value, acceptance of error
        algo_type : int
           If algo_type = 0, then simple algo will be used
        ves_type : int
           If ves_type = 1, then geo is TOROIDAL
    Returns
    =======
        kmin_vpoly : (npoly, nlos) double array
            Of the form [k_00, k_01, ..., k_0n, k_10, k_11, ..., k_1n, ...]
            where k_ij is the coefficient for the j-th flux surface
            such that the i-th ray (LOS) is closest to the extruded polygon
            at the point P_i = orig[i] + kmin[i] * vdir[i]
        dist_vpoly : (npoly, nlos) double array
            `distance[j, i]` is the distance from P_i to the i-th extruded poly.
    ---
    This is the CYTHON function, use only if you need this computation from
    cython, if you need it from Python, use `comp_dist_los_vpoly_vec`
    """
    cdef int i, ind_los, ind_pol, ind_pol2
    cdef double* loc_res
    cdef double* loc_dir
    cdef double* loc_org
    cdef double crit2, invuz,  dpar2, upar2, upscaDp
    cdef double crit2_base = eps_uz * eps_uz /400.
    cdef double** list_vpoly_x = NULL
    cdef double** list_vpoly_y = NULL
    cdef int* list_npts = NULL
    # == Tests and warnings ====================================================
    with gil:
        if not algo_type == 0 or not ves_type == 1:
            assert False
        from warnings import warn
        warn("This algo supposes that the polys are nested from inner to outer",
             Warning)
    # == Discretizing vpolys ===================================================
    list_vpoly_x = <double**>malloc(sizeof(double*)*num_poly)
    list_vpoly_y = <double**>malloc(sizeof(double*)*num_poly)
    list_npts = <int*>malloc(sizeof(int)*num_poly)
    for ind_pol in range(num_poly):
        list_vpoly_x[ind_pol] = NULL
        list_vpoly_y[ind_pol] = NULL
        _st.simple_discretize_vpoly_core(ves_poly[ind_pol],
                                        ves_poly[ind_pol].shape[1],
                                        disc_step, # discretization step
                                        &list_vpoly_x[ind_pol],
                                        &list_vpoly_y[ind_pol],
                                        &list_npts[ind_pol],
                                        0, # mode = absolute
                                        _VSMALL)
    # == Defining parallel part ================================================
    with nogil, parallel(num_threads=num_threads):
        # We use local arrays for each thread so...
        loc_dir = <double*>malloc(3*sizeof(double))
        loc_org = <double*>malloc(3*sizeof(double))
        loc_res = <double*>malloc(2*sizeof(double))
        # == The parallelization over the LOS ==================================
        for ind_los in prange(nlos, schedule='dynamic'):
            for i in range(3):
                loc_dir[i] = ray_vdir[ind_los * 3 + i]
                loc_org[i] = ray_orig[ind_los * 3 + i]
            # -- Computing values that depend on the LOS/ray -------------------
            upscaDp = loc_dir[0]*loc_org[0] + loc_dir[1]*loc_org[1]
            upar2   = loc_dir[0]*loc_dir[0] + loc_dir[1]*loc_dir[1]
            dpar2   = loc_org[0]*loc_org[0] + loc_org[1]*loc_org[1]
            invuz = 1./loc_dir[2]
            crit2 = upar2*crit2_base
            # -- Looping over each flux surface---------------------------------
            for ind_pol in range(num_poly):
                simple_dist_los_vpoly_core(loc_org, loc_dir,
                                           list_vpoly_x[ind_pol],
                                           list_vpoly_y[ind_pol],
                                           list_npts[ind_pol],
                                           upscaDp,
                                           upar2, dpar2,
                                           invuz, crit2,
                                           eps_uz, eps_vz,
                                           eps_a, eps_b,
                                           loc_res)
                if not res_k == NULL:
                    res_k[ind_los * num_poly + ind_pol] = loc_res[0]
                res_dist[ind_los * num_poly + ind_pol] = loc_res[1]
                if not loc_res[1] == loc_res[1] : #is nan
                    for ind_pol2 in range(ind_pol, num_poly):
                        res_k[ind_los * num_poly + ind_pol2] = C_NAN
                        res_dist[ind_los * num_poly + ind_pol2] = C_NAN
                    continue
        free(loc_dir)
        free(loc_org)
        free(loc_res)
    free(list_vpoly_x)
    free(list_vpoly_y)
    free(list_npts)
    return

cdef inline void simple_dist_los_vpoly_core(const double[3] ray_orig,
                                            const double[3] ray_vdir,
                                            const double* lpolyx,
                                            const double* lpolyy,
                                            const int nvert,
                                            const double upscaDp,
                                            const double upar2,
                                            const double dpar2,
                                            const double invuz,
                                            const double crit2,
                                            const double eps_uz,
                                            const double eps_vz,
                                            const double eps_a,
                                            const double eps_b,
                                            double* res_final) nogil:
    """
    This function computes the distance (and the associated k) between a Ray
    (or Line Of Sight) and an `IN` structure (a polygon extruded around the axis
    (0,0,1), eg. a flux surface).
    For more details on the algorithm please see PDF: <name_of_pdf>.pdf #TODO

    Params
    ======
        ray_orig : (3) double array
           LOS origin point coordinates, noted often : `u`
        ray_vdir : (3) double array
           LOS normalized direction vector, noted often : `D`
        lpolyx : (num_vertex) double array
           1st coordinates of the vertices of the Polygon defining the poloidal
           cut of the Vessel
        lpolyy : (num_vertex) double array
           2nd coordinates of the vertices of the Polygon defining the poloidal
           cut of the Vessel
        nvert : integer
           number of vertices describing the polygon
        upscaDp : double
           if u = [ux, uy, uz] is the direction of the ray, and D=[dx, dy, dz]
           its origin, then upscaDp = ux*dx + uy*dy
        upar2 : double
           if u = [ux, uy, uz] is the direction of the ray, and D=[dx, dy, dz]
           its origin, then upar2 = ux*ux + uy*uy
        dpar2 : double
           if u = [ux, uy, uz] is the direction of the ray, and D=[dx, dy, dz]
           its origin, then dpar2 = dx*dx + dy*dy
        invuz : double
           inverse of uz (vdir[2])
        eps_<val> : double
           Small value, acceptance of error
    Returns
    =======
        kmin_vpoly : (num_los) double array
            Of the form [k_0, k_1, ..., k_n], where k_i is the coefficient
            such that the i-th ray (LOS) is closest to the extruded polygon
            at the point P_i = orig[i] + kmin[i] * vdir[i]
        dist_vpoly : (num_los) double array
            `distance[i]` is the distance from P_i to the extruded polygon.
             if the i-th LOS intersects the poly, then distance[i] = C_NAN
    ---
    This is the cython version, only accessible from cython. If you need
    to use it from Python please use: comp_dist_los_vpoly
    """
    cdef int jj
    cdef double norm_dir2, norm_dir2_ori
    cdef double radius_z
    cdef double q = C_NAN, coeff = C_NAN, sqd = C_NAN, k = C_NAN
    cdef double v0, v1, val_a, val_b
    cdef double[2] res_a
    cdef double[2] res_b
    cdef double[3] circle_tangent
    cdef double[3] ray_vdir_norm
    cdef double rdotvec
    res_final[0] = 1000000000
    res_final[1] = 1000000000

    # == Compute all solutions =================================================
    # Set tolerance value for ray_vdir[2,ii]
    # eps_uz is the tolerated DZ across 20m (max Tokamak size)
    norm_dir2 = c_sqrt(_bgt.compute_dot_prod(ray_vdir, ray_vdir))

    norm_dir2_ori = norm_dir2
    for jj in range(3):
        ray_vdir_norm[jj] = ray_vdir[jj] / norm_dir2
    norm_dir2 = 1.

    if ray_vdir_norm[2] * ray_vdir_norm[2] < crit2:
        # -- Case with horizontal semi-line ------------------------------------
        for jj in range(nvert-1):
            if (lpolyy[jj+1] - lpolyy[jj])**2 > eps_vz * eps_vz:
                # If segment AB is NOT horizontal, then we can compute distance
                # between LOS and cone.
                # First we compute the "circle" on the cone that lives on the
                # same plane as the line
                q = (ray_orig[2] - lpolyy[jj]) / (lpolyy[jj+1] - lpolyy[jj])
                if q < 0. :
                    # Then we only need to compute distance to circle C_A
                    dist_los_circle_core(ray_vdir_norm, ray_orig,
                                        lpolyx[jj], lpolyy[jj],
                                        norm_dir2, res_a)
                elif q > 1:
                    # Then we only need to compute distance to circle C_B
                    dist_los_circle_core(ray_vdir_norm, ray_orig,
                                         lpolyx[jj+1], lpolyy[jj+1],
                                         norm_dir2, res_a)
                else:
                    # The we need to compute the radius (the height is Z_D)
                    # of the circle in the same plane as the LOS and compute the
                    # distance between the LOS and circle.
                    radius_z = q * (lpolyx[jj+1] - lpolyx[jj]) + lpolyx[jj]
                    dist_los_circle_core(ray_vdir_norm, ray_orig,
                                         radius_z, ray_orig[2],
                                         norm_dir2, res_a)
                    if res_a[1] < _VSMALL:
                        # The line is either tangent or intersects the frustum
                        # we need to make the difference
                        k = res_a[0]
                        # we compute the ray from circle center to P
                        circle_tangent[0] = -ray_orig[0] - k * ray_vdir_norm[0]
                        circle_tangent[1] = -ray_orig[1] - k * ray_vdir_norm[1]
                        circle_tangent[2] = 0. # the line is horizontal
                        rdotvec = _bgt.compute_dot_prod(circle_tangent,
                                                        ray_vdir_norm)
                        if c_abs(rdotvec) > _VSMALL:
                            # There is an intersection, distance = C_NAN
                            res_final[1] = C_NAN # distance
                            res_final[0] = C_NAN # k
                            # no need to continue
                            return
                if (res_final[1] - res_a[1] > _VSMALL
                    or (res_final[1] == res_a[1]
                        and res_final[0] - res_a[0] > _VSMALL)):
                    res_final[0] = res_a[0] # k
                    res_final[1] = res_a[1] # distance
            else:
                # -- case with horizontal cone (aka cone is a plane annulus) ---
                if (c_abs(ray_orig[2] - lpolyy[jj]) > _VSMALL
                    and c_sqrt(dpar2) >= min(lpolyx[jj], lpolyx[jj+1])
                    and c_sqrt(dpar2) <= max(lpolyx[jj], lpolyx[jj+1])):
                    # if ray and annulus are NOT on the same plane:
                    # AND the origin is somewhere in the annulus (on the
                    # X,Y plane), the origin is the closest and distance
                    # is the difference of height
                    if c_abs(ray_orig[2] - lpolyy[jj]) <= res_final[1]:
                        res_final[0] = 0 # k
                        res_final[1] = c_abs(ray_orig[2] - lpolyy[jj])
                else:
                    # Then the shortest distance is the distance to the
                    # outline circles
                    # computing dist to cricle C_A of radius R_A and height Z_A
                    dist_los_circle_core(ray_vdir_norm, ray_orig,
                                         lpolyx[jj], lpolyy[jj],
                                         norm_dir2, res_a)
                    if res_a[1] < _VSMALL:
                        # The line is either tangent or intersects the frustum
                        # we need to make the difference
                        k = res_a[0]
                        # we compute the ray from circle center to P
                        circle_tangent[0] = -ray_orig[0] - k * ray_vdir_norm[0]
                        circle_tangent[1] = -ray_orig[1] - k * ray_vdir_norm[1]
                        circle_tangent[2] = 0. # the ray is horizontal
                        rdotvec = _bgt.compute_dot_prod(circle_tangent,
                                                        ray_vdir_norm)
                        if c_abs(rdotvec) > _VSMALL:
                            # There is an intersection, distance = C_NAN
                            res_final[1] = C_NAN # distance
                            res_final[0] = C_NAN # k
                            # no need to continue
                            return
                    dist_los_circle_core(ray_vdir_norm, ray_orig,
                                         lpolyx[jj+1], lpolyy[jj+1],
                                         norm_dir2, res_b)
                    if res_b[1] < _VSMALL:
                        # The line is either tangent or intersects the frustum
                        # we need to make the difference
                        k = res_b[0]
                        # we compute the ray from circle center to P
                        circle_tangent[0] = -ray_orig[0] - k * ray_vdir_norm[0]
                        circle_tangent[1] = -ray_orig[1] - k * ray_vdir_norm[1]
                        circle_tangent[2] = 0. # the ray is horizontal
                        rdotvec = _bgt.compute_dot_prod(circle_tangent,
                                                        ray_vdir_norm)
                        if c_abs(rdotvec) > _VSMALL:
                            # There is an intersection, distance = C_NAN
                            res_final[1] = C_NAN # distance
                            res_final[0] = C_NAN # k
                            # no need to continue
                            return
                    # The result is the one associated to the shortest distance
                    if (res_final[1] - res_a[1] > _VSMALL
                        or (res_final[1] == res_a[1]
                         and res_final[0] - res_a[0] > _VSMALL)):
                        res_final[0] = res_a[0] # k
                        res_final[1] = res_a[1] # distance
                    if (res_final[1] - res_b[1] > _VSMALL
                        or (res_final[1] == res_b[1]
                         and res_final[0] - res_b[0] > _VSMALL)):
                        res_final[0] = res_b[0] # k
                        res_final[1] = res_b[1] # distance

    else:
        # == More general non-horizontal semi-line case ========================
        for jj in range(nvert-1):
            v0 = lpolyx[jj+1]-lpolyx[jj]
            v1 = lpolyy[jj+1]-lpolyy[jj]
            val_a = v0 * v0 - upar2 * v1 * invuz * v1 * invuz
            val_b = lpolyx[jj] * v0 + v1 * invuz * (
                (ray_orig[2] - lpolyy[jj]) * upar2 * invuz - upscaDp)
            coeff = - upar2 * (ray_orig[2] - lpolyy[jj])**2 * invuz * invuz +\
                    2. * upscaDp * (ray_orig[2]-lpolyy[jj]) * invuz -\
                    dpar2 + lpolyx[jj] * lpolyx[jj]
            if (val_a * val_a < eps_a * eps_a):
                if (val_b * val_b < eps_b * eps_b):
                    # let's see if C is 0 or not
                    if coeff * coeff < eps_a * eps_a :
                        # then LOS included in cone and then we can choose point
                        # such that q = 0,  k = (z_A - z_D) / uz
                        res_a[0] = (lpolyy[jj] - ray_orig[2]) * invuz
                        res_a[1] = 0 # distance = 0 since LOS in cone
                    elif v0 * v0 < eps_a and upar2 * upar2 < eps_a:
                        # cylinder and vertical line
                        if (ray_orig[2] <= max(lpolyy[jj], lpolyy[jj+1])
                            and ray_orig[2] >= min(lpolyy[jj], lpolyy[jj+1])):
                            # origin of line in the length of cylinder:
                            res_a[0] = 0
                            res_a[1] = c_abs(upar2 - lpolyx[jj])
                        elif (lpolyy[jj] >= ray_orig[2]
                              and ray_orig[2] <= lpolyy[jj+1]):
                            # ray origin below cylinder
                            if ray_vdir_norm[2] < 0:
                                # if the ray is going down origin is the closest
                                res_a[0] = 0
                                res_a[1] = c_sqrt(c_abs(
                                    (lpolyx[jj] - upar2)**2
                                    + (min(lpolyy[jj], lpolyy[jj+1]) - ray_orig[2])**2
                                ))
                            else:
                                res_a[0] = (min(lpolyy[jj], lpolyy[jj+1])
                                            - ray_orig[2]) * invuz
                                res_a[1] = c_abs(lpolyx[jj] - upar2)
                        else:
                            # ray origin above cylinder
                            if ray_vdir_norm[2] > 0:
                                # if the ray is going up origin is the closest
                                res_a[0] = 0
                                res_a[1] = c_sqrt(c_abs(
                                    (lpolyx[jj] - upar2)**2
                                    + (max(lpolyy[jj], lpolyy[jj+1]) - ray_orig[2])**2
                                ))
                            else:
                                res_a[0] = (max(lpolyy[jj], lpolyy[jj+1])
                                            - ray_orig[2]) * invuz
                                res_a[1] = c_abs(lpolyx[jj] - upar2)
                else: # (val_b * val_b > eps_b * eps_b):
                    q = -coeff / (2. * val_b)
                    if q < 0. :
                        # Then we only need to compute distance to circle C_A
                        dist_los_circle_core(ray_vdir_norm, ray_orig,
                                            lpolyx[jj], lpolyy[jj],
                                             norm_dir2, res_a)
                    elif q > 1:
                        # Then we only need to compute distance to circle C_B
                        dist_los_circle_core(ray_vdir_norm, ray_orig,
                                            lpolyx[jj+1], lpolyy[jj+1],
                                             norm_dir2, res_a)
                    else :
                        k = (q * v1 - (ray_orig[2] - lpolyy[jj])) * invuz
                        if k >= 0.:
                            # Then there is an intersection
                            res_final[0] = C_NAN
                            res_final[1] = C_NAN
                            return # no need to move forward
                        else:
                            # The closest point on the line is the LOS origin
                            res_a[0] = 0
                            res_a[1] = -k * c_sqrt(norm_dir2)

                if (res_final[1] - res_a[1] > _VSMALL
                    or (res_final[1] == res_a[1]
                        and res_final[0] - res_a[0] > _VSMALL)):
                    res_final[0] = res_a[0] # k
                    res_final[1] = res_a[1] # distance

            elif (val_b * val_b >= val_a * coeff):

                sqd = c_sqrt(c_abs(val_b * val_b - val_a * coeff))
                # First solution
                q = (-val_b + sqd) / val_a
                if q < 0:
                    # Then we only need to compute distance to circle C_A
                    dist_los_circle_core(ray_vdir_norm, ray_orig,
                                         lpolyx[jj], lpolyy[jj],
                                         norm_dir2, res_a)
                elif q > 1:
                    # Then we only need to compute distance to circle C_B
                    dist_los_circle_core(ray_vdir_norm, ray_orig,
                                         lpolyx[jj+1], lpolyy[jj+1],
                                         norm_dir2, res_a)
                else :
                    k = (q * v1 - (ray_orig[2] - lpolyy[jj])) * invuz
                    if k >= 0.:
                        # There is an intersection
                        res_final[0] = C_NAN
                        res_final[1] = C_NAN
                        return # no need to continue
                    else:
                        # The closest point on the LOS is its origin
                        res_a[0] = 0
                        res_a[1] = -k * c_sqrt(norm_dir2)

                if (res_final[1] - res_a[1] > _VSMALL
                    or (res_final[1] == res_a[1]
                        and res_final[0] - res_a[0] > _VSMALL)):
                    res_final[0] = res_a[0] # k
                    res_final[1] = res_a[1] # distance

                # Second solution
                q = (-val_b - sqd) / val_a
                if q < 0:
                    # Then we only need to compute distance to circle C_A
                    dist_los_circle_core(ray_vdir_norm, ray_orig,
                                         lpolyx[jj], lpolyy[jj],
                                         norm_dir2, res_b)
                elif q > 1:
                    # Then we only need to compute distance to circle C_B
                    dist_los_circle_core(ray_vdir_norm, ray_orig,
                                         lpolyx[jj+1], lpolyy[jj+1],
                                         norm_dir2, res_b)
                else:
                    k = (q * v1 - (ray_orig[2] - lpolyy[jj])) * invuz
                    if k>=0.:
                        # there is an intersection
                        res_final[0] = C_NAN
                        res_final[1] = C_NAN
                        return # no need to continue
                    else:
                        # The closest point on the LOS is its origin
                        res_b[0] = 0
                        res_b[1] = -k * c_sqrt(norm_dir2)

                if (res_final[1] - res_b[1] > _VSMALL
                    or (res_final[1] == res_b[1]
                        and res_final[0] - res_b[0] > _VSMALL)):
                    res_final[0] = res_b[0]
                    res_final[1] = res_b[1]

    res_final[0] = res_final[0] / norm_dir2_ori
    return

# =============================================================================
# == ARE LOS AND EXT-POLY CLOSE
# =============================================================================

cdef inline void is_close_los_vpoly_vec_core(int num_poly, int nlos,
                                             double* ray_orig,
                                             double* ray_vdir,
                                             double[:,:,::1] ves_poly,
                                             double eps_uz,
                                             double eps_a,
                                             double eps_vz,
                                             double eps_b,
                                             double eps_plane,
                                             double epsilon,
                                             int[::1] are_close,
                                             int num_threads) nogil:
    """
    This function computes the distance (and the associated k) between nlos
    Rays (or LOS) and several `IN` structures (polygons extruded around the axis
    (0,0,1), eg. flux surfaces).
    For more details on the algorithm please see PDF: <name_of_pdf>.pdf #TODO

    Params
    ======
        num_poly : int
           Number of flux surfaces
        nlos : int
           Number of LOS
        ray_orig : (3, nlos) double array
           LOS origin points coordinates
        ray_vdir : (3, nlos) double array
           LOS normalized direction vector
        ves_poly : (num_pol, 2, num_vertex) double array
           Coordinates of the vertices of the Polygon defining the 2D poloidal
           cut of the different IN surfaces.
           WARNING : we suppose all poly are nested in each other,
                     and the first one is the smallest one
        epsilon : double
           Value for testing if distance < epsilon
        eps_<val> : double
           Small value, acceptance of error
    Returns
    =======
        are_close : (npoly * num_los) bool array
            `are_close[i * num_poly + j]` indicates if distance between i-th LOS
            and j-th poly are closer than epsilon. (True if distance<epsilon)
    ---
    This is the CYTHON function, use only if you need this computation from
    cython, if you need it from Python, use `comp_dist_los_vpoly_vec`
    """
    cdef int i, ind_los, ind_pol
    cdef int npts_poly
    cdef double* loc_res
    cdef double* loc_dir
    cdef double* loc_org
    cdef double crit2, invuz,  dpar2, upar2, upscaDp
    cdef double crit2_base = eps_uz * eps_uz /400.

    # == Defining parallel part ================================================
    with nogil, parallel(num_threads=num_threads):
        # We use local arrays for each thread so...
        loc_dir = <double*>malloc(3*sizeof(double))
        loc_org = <double*>malloc(3*sizeof(double))
        loc_res = <double*>malloc(2*sizeof(double))
        # == The parallelization over the LOS ==================================
        for ind_los in prange(nlos, schedule='dynamic'):
            for i in range(3):
                loc_dir[i] = ray_vdir[ind_los * 3 + i]
                loc_org[i] = ray_orig[ind_los * 3 + i]
            # -- Computing values that depend on the LOS/ray -------------------
            upscaDp = loc_dir[0]*loc_org[0] + loc_dir[1]*loc_org[1]
            upar2   = loc_dir[0]*loc_dir[0] + loc_dir[1]*loc_dir[1]
            dpar2   = loc_org[0]*loc_org[0] + loc_org[1]*loc_org[1]
            invuz = 1./loc_dir[2]
            crit2 = upar2*crit2_base
            # -- Looping over each flux surface---------------------------------
            for ind_pol in range(num_poly):
                npts_poly = ves_poly[ind_pol].shape[1]
                simple_dist_los_vpoly_core(loc_org, loc_dir,
                                           &ves_poly[ind_pol][0][0],
                                           &ves_poly[ind_pol][1][0],
                                           npts_poly, upscaDp,
                                           upar2, dpar2,
                                           invuz, crit2,
                                           eps_uz, eps_vz,
                                           eps_a, eps_b,
                                           loc_res)
                if loc_res[1] < epsilon:
                    are_close[ind_los * num_poly + ind_pol] = 1
                elif loc_res[1] == loc_res[1]: # is nan
                    continue
        free(loc_dir)
        free(loc_org)
        free(loc_res)
    return

# ==============================================================================
# == WHICH LOS/VPOLY IS CLOSER
# ==============================================================================
cdef inline void which_los_closer_vpoly_vec_core(int num_poly, int nlos,
                                                 double* ray_orig,
                                                 double* ray_vdir,
                                                 double[:,:,::1] ves_poly,
                                                 double eps_uz,
                                                 double eps_a,
                                                 double eps_vz,
                                                 double eps_b,
                                                 double eps_plane,
                                                 int[::1] ind_close_tab,
                                                 int num_threads) nogil:
    """
    Params
    ======
        num_poly : int
           Number of flux surfaces
        nlos : int
           Number of LOS
        ray_orig : (3, nlos) double array
           LOS origin points coordinates
        ray_vdir : (3, nlos) double array
           LOS normalized direction vector
        ves_poly : (num_pol, 2, num_vertex) double array
           Coordinates of the vertices of the Polygon defining the 2D poloidal
           cut of the different IN surfaces.
           WARNING : we suppose all poly are nested in each other,
                     and the first one is the smallest one
        eps_<val> : double
           Small value, acceptance of error
    Returns
    =======
        ind_close_tab : (npoly) int array
            Of the form [ind_0, ind_1, ..., ind_(npoly-1)]
            where ind_i is the coefficient for the i-th flux surface
            such that the ind_i-th ray (LOS) is closest to the extruded polygon
            among all other LOS without going over it.
    ---
    This is the CYTHON function, use only if you need this computation from
    cython, if you need it from Python, use `comp_dist_los_vpoly_vec`
    """
    cdef int ind_los, ind_pol, indloc
    cdef double loc_dist
    cdef double* dist = <double*>malloc(sizeof(double)*num_poly*nlos)

    for indloc in range(num_poly):
        ind_close_tab[indloc] = -1
    comp_dist_los_vpoly_vec_core(num_poly, nlos,
                                 ray_orig,
                                 ray_vdir,
                                 ves_poly,
                                 eps_uz, eps_a,
                                 eps_vz, eps_b,
                                 eps_plane,
                                 True, #ves_type is Tor
                                 0, #algo_type is Simple
                                 NULL, dist,
                                 0.01,
                                 num_threads)

    # We use local arrays for each thread so...
    for ind_pol in range(num_poly):
        loc_dist = 100000000.
        for ind_los in range(nlos):
            if (dist[ind_los*num_poly + ind_pol] < loc_dist):
                ind_close_tab[ind_pol] = ind_los
                loc_dist = dist[ind_los*num_poly + ind_pol]
    free(dist)
    return

cdef inline void which_vpoly_closer_los_vec_core(int num_poly, int nlos,
                                                 double* ray_orig,
                                                 double* ray_vdir,
                                                 double[:,:,::1] ves_poly,
                                                 double eps_uz,
                                                 double eps_a,
                                                 double eps_vz,
                                                 double eps_b,
                                                 double eps_plane,
                                                 int[::1] ind_close_tab,
                                                 int num_threads) nogil:
    """
    Params
    ======
        num_poly : int
           Number of flux surfaces
        nlos : int
           Number of LOS
        ray_orig : (3, nlos) double array
           LOS origin points coordinates
        ray_vdir : (3, nlos) double array
           LOS normalized direction vector
        ves_poly : (num_pol, 2, num_vertex) double array
           Coordinates of the vertices of the Polygon defining the 2D poloidal
           cut of the different IN surfaces.
           WARNING : we suppose all poly are nested in each other,
                     and the first one is the smallest one
        eps_<val> : double
           Small value, acceptance of error
    Returns
    =======
        ind_close_los : (nlos) int array
            Of the form [ind_0, ind_1, ..., ind_(nlos-1)]
            where ind_i is the coefficient for the i-th LOS (ray)
            such that the ind_i-th poly (flux surface) is closest to the LOS
            among all other poly without going over it.
    ---
    This is the CYTHON function, use only if you need this computation from
    cython, if you need it from Python, use `comp_dist_los_vpoly_vec`
    """
    cdef int i, ind_los, ind_pol, indloc
    cdef int npts_poly
    cdef double* loc_res
    cdef double* loc_dir
    cdef double* loc_org
    cdef double crit2, invuz,  dpar2, upar2, upscaDp
    cdef double crit2_base = eps_uz * eps_uz /400.

    # initialization ...............................................
    for indloc in range(nlos):
        ind_close_tab[indloc] = num_poly-1

    # == Defining parallel part ================================================
    with nogil, parallel(num_threads=num_threads):
        # We use local arrays for each thread so...
        loc_dir = <double*>malloc(3*sizeof(double))
        loc_org = <double*>malloc(3*sizeof(double))
        loc_res = <double*>malloc(2*sizeof(double))
        # == The parallelization over the LOS ==================================
        for ind_los in prange(nlos, schedule='dynamic'):
            for i in range(3):
                loc_dir[i] = ray_vdir[ind_los * 3 + i]
                loc_org[i] = ray_orig[ind_los * 3 + i]
            # -- Computing values that depend on the LOS/ray -------------------
            upscaDp = loc_dir[0]*loc_org[0] + loc_dir[1]*loc_org[1]
            upar2   = loc_dir[0]*loc_dir[0] + loc_dir[1]*loc_dir[1]
            dpar2   = loc_org[0]*loc_org[0] + loc_org[1]*loc_org[1]
            invuz = 1./loc_dir[2]
            crit2 = upar2*crit2_base
            # -- Looping over each flux surface---------------------------------
            for ind_pol in range(num_poly):
                npts_poly = ves_poly[ind_pol].shape[1]
                simple_dist_los_vpoly_core(loc_org, loc_dir,
                                           &ves_poly[ind_pol][0][0],
                                           &ves_poly[ind_pol][1][0],
                                           npts_poly, upscaDp,
                                           upar2, dpar2,
                                           invuz, crit2,
                                           eps_uz, eps_vz,
                                           eps_a, eps_b,
                                           loc_res)
                # filling the array when nan found .............................
                if not loc_res[1] == loc_res[1]:
                    #the closer poly is the one just before
                    ind_close_tab[ind_los] = ind_pol-1
                    continue
        free(loc_dir)
        free(loc_org)
        free(loc_res)
    return




import warnings
import itertools as itt


import numpy as np


from . import _utils


# ###########################################################
# ###########################################################
#              DEFAULTS
# ###########################################################


# ###########################################################
# ###########################################################
#              Equilibrium
# ###########################################################


def main(
    din=None,
    ids=None,
    coll=None,
    key=None,
    prefix=None,
    dshort=None,
    strict=None,
    warn=None,
):

    # ------------------------
    # data as ref0 first
    # ------------------------

    lref0, dfail_ref0 = _add_ref0(
        coll=coll,
        din=din,
        dshort=dshort,
        ids=ids,
        prefix=prefix,
        strict=strict,
    )

    # ------------------------
    # data dependent only on ref0
    # ------------------------

    dfail_data_ref0 = _add_data_ref0(
        coll=coll,
        din=din,
        dshort=dshort,
        ids=ids,
        prefix=prefix,
        strict=strict,
        lref0=lref0,
    )

    # -----------------------------
    # mesh 2d second
    # -----------------------------

    dbsplines = _add_mesh_2d(
        din=din,
        coll=coll,
        ids=ids,
        prefix=prefix,
        # unused ?
        dshort=dshort,
    )

    # ---------------
    # 2d data
    # ---------------

    dfail_data2d, lk2d = _add_data_2d(
        din=din,
        ids=ids,
        coll=coll,
        dshort=dshort,
        prefix=prefix,
        strict=strict,
        # bsplines
        dbsplines=dbsplines,
    )

    # ---------------------
    # add mesh and data 1d
    # ---------------------

    dfail_data1d = _add_mesh_data_1d(
        din=din,
        ids=ids,
        coll=coll,
        dshort=dshort,
        prefix=prefix,
        strict=strict,
        # bsplines
        dbsplines=dbsplines,
        lk2d=lk2d,
    )

    # -------------
    # warnings
    # -------------

    dfail = dfail_ref0 | dfail_data_ref0 | dfail_data2d | dfail_data1d
    if warn is True and len(dfail) > 0:
        lstr = [f"\t- {k0}: {v0}" for k0, v0 in dfail.items()]
        msg = (
            "The following data could not be loaded:\n"
            f"From ids = {ids}\n"
            + "\n".join(lstr)
        )
        warnings.warn(msg)

    return


# ###########################################################
# ###########################################################
#           Ref0
# ###########################################################


def _add_ref0(
    coll=None,
    din=None,
    dshort=None,
    ids=None,
    prefix=None,
    strict=None,
):

    # --------------
    # prepare
    # --------------

    # ex: time
    lshortref = [
        k0 for k0, v0 in dshort[ids].items()
        if v0.get('ref0') is not None
    ]

    # order to start with 1d vectors
    # ex: sepR also depends on nt
    nref = [dshort[ids][k0]['long'].count('[') for k0 in lshortref]
    lshortref = [lshortref[ii] for ii in np.argsort(nref)]
    nref = [dshort[ids][k0]['long'].count('[') for k0 in lshortref]

    # initialize
    ddata = {}
    dref = {}
    dfail = {}

    # --------------
    # must have
    # --------------

    lref0 = []
    for ii, k0 in enumerate(lshortref):
        ddatai, drefi = _utils._get_short(
            din=din,
            ids=ids,
            short=k0,
            dshort=dshort,
            prefix=prefix,
            strict=strict,
        )

        # ---------------------
        # check is several ref

        if nref[ii] > 1:
            refi = [
                rr for rr in dshort[ids][k0]['ref']
                if rr not in list(drefi.keys())[0]
            ]
            assert all([rr in lref0 for rr in refi])

        # -----
        # store

        if isinstance(ddatai, dict):
            dref.update(drefi)
            ddata.update(ddatai)
            lref0.append(dshort[ids][k0]['ref0'])

        else:
            dfail[k0] = str(ddatai)

    # --------------
    # fail
    # --------------

    if len(dfail) > 0 and strict is True:
        lstr = [f"\t- {k0}: {v0}" for k0, v0 in dfail.items()]
        msg = (
            "IDS '{ids}', the following keys could not be loaded:\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    # --------------
    # store
    # --------------

    # dref
    for kr, vr in dref.items():
        coll.add_ref(**vr)

    # data
    lk = ['key', 'data', 'units', 'ref', 'dim', 'quant', 'name']
    for kd, vd in ddata.items():
        coll.add_data(
            **{k0: v0 for k0, v0 in vd.items() if k0 in lk}
        )

    return lref0, dfail


# ###########################################################
# ###########################################################
#           data ref0
# ###########################################################


def _add_data_ref0(
    coll=None,
    din=None,
    dshort=None,
    ids=None,
    prefix=None,
    strict=None,
    lref0=None,
):

    # --------------
    # prepare
    # --------------

    # ex: time
    lshort = [
        k0 for k0, v0 in dshort[ids].items()
        if v0.get('ref0') is None
        and v0.get('ref') is not None
        and all([rr in lref0 for rr in v0['ref']])
    ]

    # initialize
    ddata = {}
    dfail = {}

    # --------------
    # must have
    # --------------

    for k0 in lshort:
        ddatai, drefi = _utils._get_short(
            din=din,
            ids=ids,
            short=k0,
            dshort=dshort,
            prefix=prefix,
            strict=strict,
        )

        # -----
        # store

        if isinstance(ddatai, dict):
            assert drefi is None
            ddata.update(ddatai)

        else:
            dfail[k0] = str(ddatai)

    # --------------
    # fail
    # --------------

    if len(dfail) > 0 and strict is True:
        lstr = [f"\t- {k0}: {v0}" for k0, v0 in dfail.items()]
        msg = (
            "IDS '{ids}', the following keys could not be loaded:\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    # --------------
    # store
    # --------------

    # data
    lk = ['key', 'data', 'units', 'ref', 'dim', 'quant', 'name']
    for kd, vd in ddata.items():
        coll.add_data(
            **{k0: v0 for k0, v0 in vd.items() if k0 in lk}
        )

    return dfail

# ###########################################################
# ###########################################################
#           Mesh 2d
# ###########################################################


def _add_mesh_2d(
    din=None,
    coll=None,
    ids=None,
    prefix=None,
    # unused ?
    dshort=None,
):
    # ----------------
    # initialize
    # ----------------

    dbsplines = {}

    # ----------------
    # list of mesh 2d
    # ----------------

    if ids == 'equilibrium':

        p2d = din[ids]['time_slice'][0]['profiles_2d']
        nmesh = len(p2d)
        for im in range(nmesh):
            mtype = p2d[im]['grid_type']['name']

            if mtype == 'rectangular':
                R = p2d[im]['grid']['dim1']
                Z = p2d[im]['grid']['dim2']

                km = _utils._make_key(
                    prefix=prefix,
                    ids=ids,
                    short=f'm2d{im}',
                )

                coll.add_mesh_2d_rect(
                    key=km,
                    knots0=R,
                    knots1=Z,
                    units='m',
                    deg=1,
                )

                dbsplines[im] = f"{km}_bs1"

    return dbsplines


# ###########################################################
# ###########################################################
#           data 2d
# ###########################################################


def _add_data_2d(
    din=None,
    ids=None,
    coll=None,
    dshort=None,
    prefix=None,
    strict=None,
    # bsplines
    dbsplines=None,
):

    # ------------------
    # get list of data
    # ------------------

    if ids in ['pulse_schedule', 'summary', 'equilibrium', 'core_profiles']:
        ldata = [
            k0 for k0, v0 in dshort[ids].items()
            if '[im2d]' in v0['long']
            and 'grid' not in dshort[ids][k0]['long']
        ]

    # -------------
    # loop on data
    # -------------

    lk2d = []
    dfail = {}
    for kd in ldata:

        ddatai, drefi = _utils._get_short(
            din=din,
            ids=ids,
            short=kd,
            dshort=dshort,
            prefix=prefix,
            strict=strict,
        )

        # ----------
        # exception catching

        if not isinstance(ddatai, dict):
            dfail[kd] = ddatai
            continue

        # ------------
        # safety check

        c0 = (
            drefi is None
            and 'im2d' in ddatai[kd]['ref']
        )
        if not c0:
            msg = (
                "Inconsistent data from:\n"
                f"\t- ids = {ids}\n"
                f"\t- short = {kd}\n"
                f"\t- drefi = {drefi}\n"
                f"\t- ddatai[{kd}]['ref'] = {ddatai[kd]['ref']}\n"
            )
            raise Exception(msg)

        # ----------
        # ref

        kbs = dbsplines[0]

        ref = tuple([
            kbs if rr == 'im2d'
            else rr for rr in ddatai[kd]['ref']
        ])
        ddatai[kd]['ref'] = ref

        axis = ref.index(kbs)

        # -------------
        # safety check on shape
        # -------------

        wbs = coll._which_bsplines
        shape_bs = coll.dobj[wbs][kbs]['shape']
        shape = ddatai[kd]['data'].shape

        if shape[axis:axis+len(shape_bs)] == shape_bs:
            pass

        elif shape[axis:axis+len(shape_bs)] == shape_bs[::-1]:
            assert len(shape_bs) == 2
            ddatai[kd]['data'] = np.swapaxes(
                ddatai[kd]['data'],
                axis,
                axis + 1,
            )

        else:
            msg = (
                "2d data has unknow shape vs bsplines!\n"
                f"\t- ids = {ids}\n"
                f"\t- short = {kd}\n"
                f"\t- ref = {ref}\n"
                f"\t- kbs = {kbs}\n"
                f"\t- shape = {shape}\n"
                f"\t- coll.dobj[{wbs}][{kbs}]['shape'] = {shape_bs}\n"
            )
            raise Exception(msg)

        # -------------
        # store
        # -------------

        coll.add_data(**ddatai[kd])

        # store in list
        lk2d.append(ddatai[kd]['key'])

    # -------------
    # update ldata and complete
    # -------------

    ldata = [kk for kk in ldata if kk not in dfail.keys()]
    lpsi = [kk for kk in ldata if 'psi' in kk]

    # -----------------
    # poloidal flux

    if len(lpsi) > 0:

        kpsin = [kk for kk in lpsi if kk.endswith('psin')]
        if len(kpsin) == 0:
            pass
            # _add_psin_from_psi()

        krhopn = [kk for kk in ldata if kk.endswith('rhopn')]
        if len(krhopn) == 0:
            pass
            # _add_rhopn_from_psin()

    # -------------
    # dfail
    # -------------

    if len(dfail) > 0 and strict is True:
        lstr = [f"\t- {k0}: {str(v0)}" for k0, v0 in dfail.items()]
        msg = (
            "The following 2d data could not be loaded:\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    return dfail, lk2d


# ###########################################################
# ###########################################################
#           mesh 1d
# ###########################################################


def _add_mesh_data_1d(
    din=None,
    ids=None,
    coll=None,
    dshort=None,
    prefix=None,
    strict=None,
    # bsplines
    dbsplines=None,
    lk2d=None,
):

    # ------------------
    # get list of data
    # ------------------

    ldata = [
        k0 for k0, v0 in dshort[ids].items()
        if '[im1d]' in v0['long']
    ]

    # -------------
    # loop on data
    # -------------

    dfail = {}
    ddata = {}
    axis = None
    shape = None

    if len(ldata) == 0:
        return dfail

    # -------------
    # loop on data
    # -------------

    for kd in ldata:

        ddatai, drefi = _utils._get_short(
            din=din,
            ids=ids,
            short=kd,
            dshort=dshort,
            prefix=prefix,
            strict=strict,
        )

        # ----------
        # exception catching

        if not isinstance(ddatai, dict):
            dfail[kd] = ddatai
            continue

        # ------------
        # safety check

        c0 = (
            drefi is None
            and np.sum(['im1d' in rr for rr in ddatai[kd]['ref']])
        )
        if not c0:
            msg = (
                "Inconsistent data from:\n"
                f"\t- ids = {ids}\n"
                f"\t- short = {kd}\n"
                f"\t- drefi = {drefi}\n"
                f"\t- ddatai[{kd}]['ref'] = {ddatai[kd]['ref']}\n"
            )
            raise Exception(msg)

        # shape
        shapei = ddatai[kd]['data'].shape
        if shape is None:
            shape = shapei
        else:
            assert shapei == shape

        # axis
        r1d = [rr for rr in ddatai[kd]['ref'] if 'm1d' in rr][0]
        axisi = ddatai[kd]['ref'].index(r1d)
        if axis is None:
            axis = axisi
        else:
            assert axis == axisi

        # ------------------
        # aggregate to ddata

        ddata[kd] = ddatai[kd]

    # update ldata
    ldata = [kk for kk in ldata if kk in ddata.keys()]

    # -------------------------------
    # identify 1d knots and 2d subkey
    # -------------------------------

    k1d, q1d, k2d = _get_subkey(
        coll=coll,
        ids=ids,
        shape=shape,
        axis=axis,
        ddata=ddata,
        ldata=ldata,
        lk2d=lk2d,
    )

    # --------------------
    # add mesh
    # --------------------

    km = _utils._make_key(
        prefix=prefix,
        ids=ids,
        short='m1d',
    )

    lk = ['units', 'dim', 'quant', 'name']
    coll.add_mesh_1d(
        key=km,
        knots=q1d,
        subkey=k2d,
        deg=1,
        **{k0: ddata[k1d].get(k0) for k0 in lk},
    )

    kbs = f"{km}_bs1"

    # ----------------
    # add data 1d
    # ----------------

    for kd in ldata:

        if kd == k1d:
            continue

        # ----------
        # ref

        ref = tuple([
            kbs if 'im1d' in rr
            else rr for rr in ddata[kd]['ref']
        ])
        ddata[kd]['ref'] = ref

        axis = ref.index(kbs)

        # -------------
        # store
        # -------------

        coll.add_data(**ddata[kd])

    # -------------
    # dfail
    # -------------

    if len(dfail) > 0 and strict is True:
        lstr = [f"\t- {k0}: {str(v0)}" for k0, v0 in dfail.items()]
        msg = (
            "The following 1d data could not be loaded:\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    return dfail


# ################################################################
# ################################################################
#        Identify subkey
# ################################################################


def _get_subkey(
    coll=None,
    ids=None,
    shape=None,
    axis=None,
    ddata=None,
    ldata=None,
    lk2d=None,
):

    # ----------------
    # identify mesh 1d
    # ----------------

    sli = tuple([
        slice(None) if ii == axis
        else slice(0, 1, 1)
        for ii in range(len(shape))
    ])

    l1d = [
        k0 for k0, v0 in ddata.items()
        if np.allclose(v0['data'], v0['data'][sli])
    ]
    if len(l1d) != 1:
        msg = (
            "No / multiple constant 1d mesh data identified:\n"
            f"\t- ids = {ids}\n"
            f"\t- l1d = {l1d}\n"
        )
        raise Exception(msg)

    k1d = l1d[0]
    q1d = ddata[l1d[0]]['data'][sli].ravel()

    # ------------------
    # Identify 2d subkey
    # ------------------

    k2d = [kk for kk in lk2d if ddata[k1d]['name'] == coll.ddata[kk]['name']]
    if len(k2d) > 1:
        msg = (
            "Several 2d data identified to match 1d mesh:\n"
            "\t- ids = {ids}\n"
            "\t- k1d = {k1d}\n"
            "\t- k2d = {k2d}\n"
        )
        raise Exception(msg)

    # --------------------
    # no match => add psin
    # --------------------

    if len(k2d) == 0 and k1d[-1] == 'n':

        # get k2
        lk1 = [kk for kk in ldata if kk.endswith(k1d[:-1])]
        lk2 = [kk for kk in lk2d if kk.endswith(k1d.replace('1d', '2d')[:-1])]
        if len(lk1) != 1:
            msg = f"Unidentified 1d base for 2d subkey: {lk1}"
            raise Exception(msg)
        if len(lk2) != 1:
            msg = f"Unidentified 2d subkey: {lk2}"
            raise Exception(msg)
        k2d = lk2[0]

        # slices
        sli0 = tuple([
            0 if ii == axis else slice(None) for ii in range(len(shape))
        ])
        sli1 = tuple([
            -1 if ii == axis else slice(None) for ii in range(len(shape))
        ])
        sli2 = tuple(itt.chain.from_iterable([
            [None, None] if ii == axis
            else [slice(None)] for ii in range(len(shape))
        ]))

        # compute normalization
        q1d0 = ddata[lk1[0]]['data'][sli0][sli2]
        q1d1 = ddata[lk1[0]]['data'][sli1][sli2]
        q2dn = (coll.ddata[k2d]['data'] - q1d0) / (q1d1 - q1d0)

        # add data2d
        k2dn = f"{k2d}n"
        coll.add_data(
            key=k2dn,
            data=q2dn,
            ref=coll.ddata[k2d]['ref'],
        )

    else:
        k2dn = k2d[0]

    return k1d, q1d, k2dn
